"""Functionailty for drawing tensor networks."""

import collections
import importlib
import textwrap
import warnings

import numpy as np

from ..utils import autocorrect_kwargs, check_opt, valmap

# from ..schematic import average_color, darken_color, auto_colors, hash_to_color

HAS_FA2 = importlib.util.find_spec("fa2") is not None


@autocorrect_kwargs
def draw_tn(
    tn,
    color=None,
    *,
    show_inds=None,
    show_tags=None,
    output_inds=None,
    highlight_inds=(),
    highlight_tids=(),
    highlight_inds_color=(1.0, 0.2, 0.2),
    highlight_tids_color=(1.0, 0.2, 0.2),
    custom_colors=None,
    legend="auto",
    dim=2,
    fix=None,
    layout="auto",
    initial_layout="auto",
    refine_layout="auto",
    iterations="auto",
    k=None,
    pos=None,
    node_color=None,
    node_scale=1.0,
    node_size=None,
    node_alpha=1.0,
    node_shape="o",
    node_outline_size=None,
    node_outline_darkness=0.9,
    node_hatch="",
    edge_color=None,
    edge_scale=1.0,
    edge_alpha=1 / 2,
    multi_edge_spread=0.1,
    multi_tag_style="auto",
    show_left_inds=True,
    arrow_opts=None,
    label_color=None,
    font_size=10,
    font_size_inner=7,
    font_family="monospace",
    isdark=None,
    title=None,
    backend="matplotlib",
    figsize=(6, 6),
    margin=None,
    xlims=None,
    ylims=None,
    adjust_lims=True,
    get=None,
    return_fig=False,
    ax=None,
):
    """Plot this tensor network as a networkx graph using matplotlib,
    with edge width corresponding to bond dimension.

    Parameters
    ----------
    color : sequence of tags or True, optional
        If given, color all tensor with each of these tags with a unique color.
        If `True`, do this for all tags in the tensor network.
    output_inds : sequence of str, optional
        For hyper tensor networks explicitly specify which indices should be
        drawn as outer indices. If not set, the outer indices are assumed to be
        those that only appear on a single tensor.
    highlight_inds : iterable, optional
        Highlight these edges.
    highlight_tids : iterable, optional
        Highlight these nodes.
    highlight_inds_color
        What color to use for ``highlight_inds`` nodes.
    highlight_tids_color : tuple[float], optional
        What color to use for ``highlight_tids`` nodes.
    show_inds : {None, False, True, 'all', 'bond-size'}, optional
        Explicitly turn on labels for each tensors indices.
    show_tags : {None, False, True}, optional
        Explicitly turn on labels for each tensors tags.
    custom_colors : sequence of colors, optional
        Supply a custom sequence of colors to match the tags given
        in ``color``.
    title : str, optional
        Set a title for the axis.
    legend : "auto" or bool, optional
        Whether to draw a legend for the colored tags. If ``"auto"`` then
        only draw a legend if there are less than 20 tags.
    dim : {2, 2.5, 3}, optional
        What dimension to position the graph nodes in. 2.5 positions the nodes
        in 3D but then projects then down to 2D.
    fix : dict[tags_ind_or_tid], (float, float)], optional
        Used to specify actual relative positions for each tensor node.
        Each key should be a sequence of tags that uniquely identifies a
        tensor, a ``tid``, or a ``ind``, and each value should be a ``(x, y)``
        coordinate tuple.
    layout : str, optional
        How to layout the graph. Can be any of the following:

            - ``'auto'``: layout the graph using a networkx method then relax
              the layout using a force-directed algorithm.
            - a networkx layout method name, e.g. ``'kamada_kawai'``: just
              layout the graph using a networkx method, with no relaxation.
            - a graphviz method such as ``'dot'``, ``'neato'`` or ``'sfdp'``:
              layout the graph using ``pygraphviz``.

    initial_layout : {'auto', 'spectral', 'kamada_kawai', 'circular', \\
                      'planar', 'random', 'shell', 'bipartite', ...}, optional
        If ``layout == 'auto'`` The name of a networkx layout to use before
        iterating with the spring layout. Set `layout` directly or
        ``iterations=0`` if you don't want any spring relaxation.
    iterations : int, optional
        How many iterations to perform when when finding the best layout
        using node repulsion. Ramp this up if the graph is drawing messily.
    k : float, optional
        The optimal distance between nodes.
    pos : dict, optional
        Pre-computed positions for the nodes. If given, this will override
        ``layout``. The nodes shouuld be exactly the same as the nodes in the
        graph returned by ``draw(get='graph')``.
    node_color : tuple[float], optional
        Default color of nodes.
    node_scale : float, optional
        Scale the node sizes by this factor, in addition to the automatic
        scaling based on the number of tensors.
    node_size : None, float or dict, optional
        How big to draw the tensors. Can be a global single value, or a dict
        containing values for specific tags or tids. This is in absolute
        figure units. See ``node_scale`` simply scale the node sizes up or
        down.
    node_alpha : float, optional
        Transparency of the nodes.
    node_shape : None, str or dict, optional
        What shape to draw the tensors. Should correspond to a matplotlib
        scatter marker. Can be a global single value, or a dict containing
        values for specific tags or tids.
    node_outline_size : None, float or dict, optional
        The width of the border of each node. Can be a global single value, or
        a dict containing values for specific tags or tids.
    node_outline_darkness : float, optional
        Darkening of nodes outlines.
    edge_color : tuple[float], optional
        Default color of edges.
    edge_scale : float, optional
        How much to scale the width of the edges.
    edge_alpha : float, optional
        Set the alpha (opacity) of the drawn edges.
    multi_edge_spread : float, optional
        How much to spread the lines of multi-edges.
    show_left_inds : bool, optional
        Whether to show ``tensor.left_inds`` as incoming arrows.
    arrow_closeness : float, optional
        How close to draw the arrow to its target.
    arrow_length : float, optional
        The size of the arrow with respect to the edge.
    arrow_overhang : float, optional
        Varies the arrowhead between a triangle (0.0) and 'V' (1.0).
    arrow_linewidth : float, optional
        The width of the arrow line itself.
    label_color : tuple[float], optional
        Color to draw labels with.
    font_size : int, optional
        Font size for drawing tags and outer indices.
    font_size_inner : int, optional
        Font size for drawing inner indices.
    font_family : str, optional
        Font family to use for all labels.
    isdark : bool, optional
        Explicitly specify that the background is dark, and use slightly
        different default drawing colors. If not specified detects
        automatically from `matplotlib.rcParams`.
    figsize : tuple of int, optional
        The size of the drawing.
    margin : None or float, optional
        Specify an argument for ``ax.margin``, else the plot limits will try
        and be computed based on the node positions and node sizes.
    xlims : None or tuple, optional
        Explicitly set the x plot range.
    xlims : None or tuple, optional
        Explicitly set the y plot range.
    adjust_lims : bool or "auto", optional
        Whether to automatically adjust the axes limits as elements are added.
        If "auto" (the default), then this is ``True`` if ``ax`` is None, else
        ``False``, i.e. only adjust limits if we own the figure.
    get : {None, 'pos', 'graph'}, optional
        If ``None`` then plot as normal, else if:

        - ``'pos'``, return the plotting positions of each ``tid`` and ``ind``
          drawn as a node, this can supplied to subsequent calls as ``fix=pos``
          to maintain positions, even as the graph structure changes.
        - ``'graph'``, return the ``networkx.Graph`` object. Note that this
          will potentially have extra nodes representing output and hyper
          indices.

    return_fig : bool, optional
        If True and ``ax is None`` then return the figure created rather than
        executing ``pyplot.show()``.
    ax : matplotlib.Axis, optional
        Draw the graph on this axis rather than creating a new figure.
    """
    import math

    import matplotlib as mpl
    import networkx as nx
    from matplotlib.colors import to_rgb, to_rgba

    from ..schematic import darken_color, hash_to_color

    check_opt(
        "multi_tag_style",
        multi_tag_style,
        ("auto", "pie", "nest", "average", "last"),
    )

    if output_inds is None:
        output_inds = set(tn.outer_inds())
    elif isinstance(output_inds, str):
        output_inds = {output_inds}
    else:
        output_inds = set(output_inds)

    # automatically decide whether to show tags and inds
    if show_inds is None:
        show_inds = len(tn.outer_inds()) <= 20
    show_inds = {False: "", True: "outer"}.get(show_inds, show_inds)

    if show_tags is None:
        show_tags = len(tn.tag_map) <= 20
    show_tags = {False: "", True: "tags"}.get(show_tags, show_tags)

    if isdark is None:
        isdark = sum(to_rgb(mpl.rcParams["figure.facecolor"])) / 3 < 0.5

    if isdark:
        default_draw_color = (0.55, 0.57, 0.60, 1.0)
        default_label_color = (0.85, 0.86, 0.87, 1.0)
    else:
        default_draw_color = (0.45, 0.47, 0.50, 1.0)
        default_label_color = (0.33, 0.34, 0.35, 1.0)

    if edge_color is None:
        edge_color = mpl.colors.to_rgba(default_draw_color, edge_alpha)
    elif edge_color is True:
        # hash edge to get color
        pass
    else:
        edge_color = mpl.colors.to_rgba(edge_color, edge_alpha)

    if node_color is None:
        node_color = mpl.colors.to_rgba(default_draw_color, node_alpha)
    else:
        node_color = mpl.colors.to_rgba(node_color, node_alpha)

    if label_color is None:
        label_color = default_label_color
    elif label_color == "inherit":
        label_color = mpl.rcParams["axes.labelcolor"]

    # get colors for tagged nodes
    if color is True:
        # color all tags automatically
        color = tuple(tn.tag_map.keys())

    colors = get_colors(color, custom_colors, node_alpha)

    if legend == "auto":
        legend = len(colors) <= 20

    highlight_tids_color = to_rgba(highlight_tids_color, node_alpha)
    highlight_inds_color = to_rgba(highlight_inds_color, edge_alpha)

    # set the size of the nodes and their border
    node_size = parse_dict_to_tids_or_inds(
        node_size,
        tn,
        default=1,
    )
    node_outline_size = parse_dict_to_tids_or_inds(
        node_outline_size,
        tn,
        default=1,
    )
    node_shape = parse_dict_to_tids_or_inds(node_shape, tn, default="o")
    node_hatch = parse_dict_to_tids_or_inds(node_hatch, tn, default="")

    # build the graph
    edges = collections.defaultdict(lambda: collections.defaultdict(list))
    nodes = collections.defaultdict(dict)

    # parse all indices / edges
    for ix, tids in tn.ind_map.items():
        tids = sorted(tids)

        isouter = ix in output_inds
        ishyper = isouter or (len(tids) != 2)
        ind_size = tn.ind_size(ix)
        edge_size = edge_scale * math.log2(ind_size)

        # compute a color for this index
        color = (
            highlight_inds_color
            if ix in highlight_inds
            else to_rgba(hash_to_color(ix))
            if edge_color is True
            else edge_color
        )

        # compute a label for this index
        if ishyper:
            # each tensor connects to the dummy node represeting the hyper edge
            pairs = [(tid, ix) for tid in tids]
            dummy_nodes = [ix]
            if isouter and len(tids) > 1:
                # 'hyper outer' index
                pairs.append((("outer", ix), ix))
                dummy_nodes.append(("outer", ix))
            # hyper labels get put on dummy node
            label = ""

            for dnode in dummy_nodes:
                nodes[dnode]["ind"] = ix
                nodes[dnode]["ind_size"] = ind_size
                # make actual node invisible
                nodes[dnode]["color"] = (1.0, 1.0, 1.0, 1.0)
                nodes[dnode]["size"] = 0.0
                nodes[dnode]["outline_size"] = 0.0
                nodes[dnode]["outline_color"] = (1.0, 1.0, 1.0, 1.0)
                nodes[dnode]["marker"] = "."  # set to avoid warning, size=0
                nodes[dnode]["hatch"] = ""

                # set these for plotly hover info
                nodes[dnode]["tid"] = nodes[ix]["shape"] = nodes[ix][
                    "tags"
                ] = ""

                should_label = len(dummy_nodes) == 1 or isinstance(
                    dnode, tuple
                )
                # put the labels on this dummy node
                # -> either at center of inner hyper index
                # -> of danling on outer hyper index

                if (
                    should_label
                    and ((show_inds == "outer") and isouter)
                    or (show_inds == "all")
                ):
                    # show as outer index or inner index name
                    nodes[dnode]["label"] = ix
                elif should_label and show_inds == "bond-size":
                    # show all bond sizes
                    nodes[dnode]["label"] = f"{tn.ind_size(ix)}"
                else:
                    # labels hidden or inner edge
                    nodes[dnode]["label"] = ""

                nodes[dnode]["label_fontsize"] = font_size_inner
                nodes[dnode]["label_color"] = label_color
                nodes[dnode]["label_fontfamily"] = font_family

        else:
            # standard edge
            pairs = [tuple(tids)]

            if show_inds == "all":
                # show inner index name
                label = ix
            elif show_inds == "bond-size":
                # show all bond sizes
                label = f"{ind_size}"
            else:
                # labels hidden or inner edge
                label = ""

        for pair in pairs:
            edges[pair]["color"].append(color)
            edges[pair]["ind"].append(ix)
            edges[pair]["ind_size"].append(ind_size)
            edges[pair]["edge_size"].append(edge_size)
            edges[pair]["label"].append(label)
            edges[pair]["label_fontsize"] = font_size_inner
            edges[pair]["label_color"] = label_color
            edges[pair]["label_fontfamily"] = font_family

            if isinstance(pair[0], tuple):
                # dummy hyper outer edge - no arrows
                edges[pair]["arrow_left"].append(False)
                edges[pair]["arrow_right"].append(False)
                edges[pair]["label_left"].append(None)
                edges[pair]["label_right"].append(None)
            else:
                # tensor side can always have an incoming arrow
                tl = tn.tensor_map[pair[0]]
                tl_left_inds = tl.left_inds
                edges[pair]["arrow_left"].append(
                    show_left_inds
                    and (tl_left_inds is not None)
                    and (ix in tl_left_inds)
                )

                if hasattr(tl.data, "signature"):
                    sigl = tl.data.signature[tl.inds.index(ix)]
                    edges[pair]["label_left"].append(sigl)
                else:
                    edges[pair]["label_left"].append(None)

                if ishyper:
                    # hyper edge can't have an incoming arrow
                    edges[pair]["arrow_right"].append(False)
                    edges[pair]["label_right"].append(None)
                else:
                    # standard edge can
                    tr = tn.tensor_map[pair[1]]
                    tr_left_inds = tr.left_inds
                    edges[pair]["arrow_right"].append(
                        show_left_inds
                        and (tr_left_inds is not None)
                        and (ix in tr_left_inds)
                    )

                    if hasattr(tr.data, "signature"):
                        sigr = tr.data.signature[tr.inds.index(ix)]
                        edges[pair]["label_right"].append(sigr)
                    else:
                        edges[pair]["label_right"].append(None)

    # parse all tensors / nodes
    for tid, t in tn.tensor_map.items():
        nodes[tid]["tid"] = tid
        nodes[tid]["tags"] = str(list(t.tags))
        nodes[tid]["shape"] = str(t.shape)
        nodes[tid]["size"] = node_size[tid]
        nodes[tid]["outline_size"] = node_outline_size[tid]
        nodes[tid]["marker"] = node_shape[tid]
        nodes[tid]["hatch"] = node_hatch[tid]

        if show_tags == "tags":
            node_label = ", ".join(map(str, t.tags))
            # make the tags appear with auto vertical extent
            nodes[tid]["label"] = "\n".join(
                textwrap.wrap(node_label, max(2 * len(node_label) ** 0.5, 16))
            )
        elif show_tags == "tids":
            nodes[tid]["label"] = str(tid)
        elif show_tags == "shape":
            nodes[tid]["label"] = nodes[tid]["shape"]
        else:
            nodes[tid]["label"] = ""

        nodes[tid]["label_fontsize"] = font_size
        nodes[tid]["label_color"] = label_color
        nodes[tid]["label_fontfamily"] = font_family

        if tid in highlight_tids:
            nodes[tid]["color"] = highlight_tids_color
            nodes[tid]["outline_color"] = darken_color(
                highlight_tids_color, node_outline_darkness
            )
        else:
            # collect all relevant tag colors
            multi_colors = []
            multi_outline_colors = []
            for tag in colors:
                if tag in t.tags:
                    multi_colors.append(colors[tag])
                    multi_outline_colors.append(
                        darken_color(colors[tag], node_outline_darkness)
                    )

            if len(multi_colors) >= 1:
                # set the basic color to the last tag
                nodes[tid]["color"] = multi_colors[-1]
                nodes[tid]["outline_color"] = multi_outline_colors[-1]
                if len(multi_colors) >= 2:
                    # have multiple relevant tags - store them, but some
                    # backends might support, so store alongside basic color
                    nodes[tid]["multi_colors"] = multi_colors
                    nodes[tid]["multi_outline_colors"] = multi_outline_colors
            else:
                # untagged node
                nodes[tid]["color"] = node_color
                nodes[tid]["outline_color"] = darken_color(
                    node_color, node_outline_darkness**2
                )

    G = nx.Graph()
    for edge, edge_data in edges.items():
        G.add_edge(*edge, **edge_data)
    for node, node_data in nodes.items():
        G.add_node(node, **node_data)

    if pos is None:
        pos = get_positions(
            tn=tn,
            G=G,
            fix=fix,
            layout=layout,
            initial_layout=initial_layout,
            refine_layout=refine_layout,
            k=k,
            dim=dim,
            iterations=iterations,
        )
    else:
        pos = _normalize_positions(pos)

    if tn.num_tensors > 0:
        # compute a base size using the position and number of tensors
        # first get plot volume:
        node_packing_factor = tn.num_tensors**-0.45
        xs, ys, *zs = zip(*pos.values())
        xmin, xmax = min(xs), max(xs)
        ymin, ymax = min(ys), max(ys)
        # if there only a few tensors we don't want to limit the node size
        # because of flatness, also don't allow the plot volume to go to zero
        xrange = max(((xmax - xmin) / 2, node_packing_factor, 0.1))
        yrange = max(((ymax - ymin) / 2, node_packing_factor, 0.1))
        plot_volume = xrange * yrange
        if zs:
            zmin, zmax = min(zs[0]), max(zs[0])
            zrange = max(((zmax - zmin) / 2, node_packing_factor, 0.1))
            plot_volume *= zrange
        # in total we account for:
        #     - user specified scaling
        #     - number of tensors
        #     - how flat the plot area is (flatter requires smaller nodes)
        full_node_scale = (
            0.2 * node_scale * node_packing_factor * plot_volume**0.5
        )

    else:
        full_node_scale = 1.0

    default_outline_size = 6 * full_node_scale**0.5

    # update node size and position attributes
    for node, node_data in nodes.items():
        nodes[node]["size"] = G.nodes[node]["size"] = (
            full_node_scale * node_data["size"]
        )
        nodes[node]["outline_size"] = G.nodes[node]["outline_size"] = (
            default_outline_size * node_data["outline_size"]
        )
        nodes[node]["coo"] = G.nodes[node]["coo"] = pos[node]

    for (i, j), edge_data in edges.items():
        edge_data["coos"] = G.edges[i, j]["coos"] = pos[i], pos[j]
        edge_data["shorten"] = G.edges[i, j]["shorten"] = (
            nodes[i]["size"],
            nodes[j]["size"],
        )

    if get == "pos":
        return pos
    if get == "graph,pos":
        return G, pos

    opts = {
        "colors": colors,
        "node_outline_darkness": node_outline_darkness,
        "title": title,
        "legend": legend,
        "multi_edge_spread": multi_edge_spread,
        "multi_tag_style": multi_tag_style,
        "arrow_opts": arrow_opts,
        "label_color": label_color,
        "font_family": font_family,
        "figsize": figsize,
        "margin": margin,
        "xlims": xlims,
        "ylims": ylims,
        "return_fig": return_fig,
        "ax": ax,
    }

    if get == "data":
        return edges, nodes, opts

    if backend == "matplotlib":
        return _draw_matplotlib(
            edges=edges,
            nodes=nodes,
            adjust_lims=adjust_lims,
            **opts,
        )

    if backend == "matplotlib3d":
        return _draw_matplotlib3d(G, **opts)

    if backend == "plotly":
        return _draw_plotly(G, **opts)


def parse_dict_to_tids_or_inds(spec, tn, default="__NONE__"):
    """Parse a dictionary possibly containing a mix of tags, tids and inds, to
    a dictionary with only sinlge tids and inds as keys. If a tag or set of
    tags are given as a key, all matching tensor tids will receive the value.
    """
    #
    if (spec is not None) and (not isinstance(spec, dict)):
        # assume new default value for everything
        return collections.defaultdict(lambda: spec)

    # allow not specifying a default value
    if default != "__NONE__":
        new = collections.defaultdict(lambda: default)
    else:
        new = {}

    if spec is None:
        return new

    # parse the special values
    for k, v in spec.items():
        if (
            # given as tid
            (isinstance(k, int) and k in tn.tensor_map)
            or
            # given as ind
            (isinstance(k, str) and k in tn.ind_map)
        ):
            # already a tid
            new[k] = v
            continue

        try:
            for tid in tn._get_tids_from_tags(k):
                new[tid] = v
        except KeyError:
            # just ignore keys that don't match any tensor
            pass

    return new


def _add_legend_matplotlib(
    ax, colors, legend, node_outline_darkness, label_color, font_family
):
    import matplotlib.pyplot as plt

    # create legend
    if colors and legend:
        handles = []
        for color in colors.values():
            ecolor = tuple(
                (1.0 if i == 3 else node_outline_darkness) * c
                for i, c in enumerate(color)
            )
            handles += [
                plt.Line2D(
                    [0],
                    [0],
                    marker="o",
                    color=color,
                    markeredgecolor=ecolor,
                    markeredgewidth=1,
                    linestyle="",
                    markersize=10,
                )
            ]

        # needed in case '_' is the first character
        lbls = [f" {lbl}" for lbl in colors]

        legend = ax.legend(
            handles,
            lbls,
            ncol=max(round(len(handles) / 20), 1),
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            labelcolor=label_color,
            prop={"family": font_family},
        )
        # do this manually as otherwise can't make only face transparent
        legend.get_frame().set_alpha(None)
        legend.get_frame().set_facecolor((0.0, 0.0, 0.0, 0.0))
        legend.get_frame().set_edgecolor((0.6, 0.6, 0.6, 0.2))


def _draw_matplotlib(
    edges,
    nodes,
    *,
    colors=None,
    node_outline_darkness=0.9,
    title=None,
    legend=True,
    multi_edge_spread=0.1,
    multi_tag_style="auto",
    arrow_opts=None,
    label_color=None,
    font_family="monospace",
    figsize=(6, 6),
    margin=None,
    xlims=None,
    ylims=None,
    adjust_lims=True,
    return_fig=False,
    ax=None,
):
    import matplotlib.pyplot as plt

    from quimb.schematic import Drawing, average_color

    d = Drawing(figsize=figsize, ax=ax, adjust_lims=adjust_lims)
    if ax is None:
        fig = d.fig
        ax = d.ax
        fig.patch.set_alpha(0.0)
        ax.patch.set_alpha(0.0)
    else:
        fig = None

    arrow_opts = arrow_opts or {}
    arrow_opts.setdefault("center", 0.8)
    arrow_opts.setdefault("linewidth", 1)
    arrow_opts.setdefault("width", 0.08)
    arrow_opts.setdefault("length", 0.12)

    if title is not None:
        ax.set_title(str(title), color=label_color)

    for _, edge_data in edges.items():
        cooa, coob = edge_data["coos"]
        edge_colors = edge_data["color"]
        edge_sizes = edge_data["edge_size"]
        labels = edge_data["label"]
        arrow_lefts = edge_data["arrow_left"]
        arrow_rights = edge_data["arrow_right"]
        label_lefts = edge_data["label_left"]
        label_rights = edge_data["label_right"]
        multiplicity = len(edge_colors)
        shorten = edge_data["shorten"]

        if multiplicity > 1:
            offsets = np.linspace(
                +multiplicity * multi_edge_spread / 2,
                -multiplicity * multi_edge_spread / 2,
                multiplicity,
            )
        else:
            offsets = None

        for m in range(multiplicity):
            line_opts = dict(
                cooa=cooa,
                coob=coob,
                linewidth=edge_sizes[m],
                color=edge_colors[m],
            )

            arrowhead, reverse = {
                (False, False): (None, False),  # no arrow
                (False, True): (True, False),  # arrowhead to right
                (True, False): (True, True),  # arrowhead to left
                (True, True): (True, "both"),  # arrowheads both sides
            }[arrow_lefts[m], arrow_rights[m]]

            if arrowhead:
                line_opts["arrowhead"] = dict(
                    reverse=reverse,
                    **arrow_opts,
                )

            if labels[m]:
                line_opts["text"] = dict(
                    text=labels[m],
                    fontsize=edge_data["label_fontsize"],
                    color=edge_data["label_color"],
                    fontfamily=edge_data["label_fontfamily"],
                )
            if label_lefts[m]:
                line_opts["text_left"] = dict(
                    text=label_lefts[m],
                    fontsize=edge_data["label_fontsize"] + 3,
                    color=edge_data["label_color"],
                    fontfamily=edge_data["label_fontfamily"],
                )
            if label_rights[m]:
                line_opts["text_right"] = dict(
                    text=label_rights[m],
                    fontsize=edge_data["label_fontsize"] + 3,
                    color=edge_data["label_color"],
                    fontfamily=edge_data["label_fontfamily"],
                )

            if multiplicity > 1:
                d.line_offset(offset=offsets[m], **line_opts)
            else:
                d.line(shorten=shorten, **line_opts)

    # draw the tensors
    for _, node_data in nodes.items():
        patch_opts = dict(
            coo=node_data["coo"],
            radius=node_data["size"],
            facecolor=node_data["color"],
            edgecolor=node_data["outline_color"],
            linewidth=node_data["outline_size"],
            hatch=node_data["hatch"],
        )
        marker = node_data["marker"]

        if "multi_colors" in node_data:
            # tensor has multiple tags which are colored

            if multi_tag_style in ("pie", "auto"):
                # draw a mini pie chart
                if marker not in ("o", "."):
                    warnings.warn(
                        "Can only draw multi-colored nodes as circles."
                    )

                angles = np.linspace(
                    0, 360, len(node_data["multi_colors"]) + 1
                )
                for i, (color, outline_color) in enumerate(
                    zip(
                        node_data["multi_colors"],
                        node_data["multi_outline_colors"],
                    )
                ):
                    patch_opts["facecolor"] = color
                    patch_opts["edgecolor"] = outline_color
                    d.wedge(
                        theta1=angles[i] - 67.5,
                        theta2=angles[i + 1] - 67.5,
                        **patch_opts,
                    )
            elif multi_tag_style == "nest":
                # draw nested markers of decreasing size
                radii = np.linspace(
                    node_data["size"], 0, len(node_data["multi_colors"]) + 1
                )
                for i, (color, outline_color) in enumerate(
                    zip(
                        node_data["multi_colors"],
                        node_data["multi_outline_colors"],
                    )
                ):
                    patch_opts["facecolor"] = color
                    patch_opts["edgecolor"] = outline_color
                    d.marker(
                        marker=marker,
                        **{**patch_opts, "radius": radii[i], "linewidth": 0},
                    )
            elif multi_tag_style == "last":
                # draw a single marker with last tag
                patch_opts["facecolor"] = node_data["multi_colors"][-1]
                patch_opts["edgecolor"] = node_data["multi_outline_colors"][-1]
                d.marker(marker=marker, **patch_opts)
            else:  # multi_tag_style == "average":
                # draw a single marker with average color
                patch_opts["facecolor"] = average_color(
                    node_data["multi_colors"]
                )
                patch_opts["edgecolor"] = average_color(
                    node_data["multi_outline_colors"]
                )
                d.marker(marker=marker, **patch_opts)

        else:
            d.marker(marker=marker, **patch_opts)

        if node_data["label"]:
            d.text(
                node_data["coo"],
                node_data["label"],
                fontsize=node_data["label_fontsize"],
                color=node_data["label_color"],
                fontfamily=node_data["label_fontfamily"],
            )

    _add_legend_matplotlib(
        ax, colors, legend, node_outline_darkness, label_color, font_family
    )

    if fig is None:
        # ax was supplied, don't modify and simply return
        return
    else:
        # axes and figure were created
        if xlims is not None:
            ax.set_xlim(xlims)
        if ylims is not None:
            ax.set_ylim(ylims)
        if margin is not None:
            ax.margins(margin)

    if return_fig:
        return fig
    else:
        plt.show()
        plt.close(fig)


def _linearize_graph_data(G, multi_tag_style="auto"):
    from ..schematic import average_color

    edge_source = collections.defaultdict(list)
    for _, _, edge_data in G.edges(data=True):
        cooa, coob = edge_data["coos"]
        x0, y0, *maybe_z0 = cooa
        x1, y1, *maybe_z1 = coob
        edge_source["x0"].append(x0)
        edge_source["y0"].append(y0)
        edge_source["x1"].append(x1)
        edge_source["y1"].append(y1)
        if maybe_z0:
            edge_source["z0"].extend(maybe_z0)
            edge_source["z1"].extend(maybe_z1)

        # we just aggregate all multi-edges into one
        edge_source["color"].append(average_color(edge_data["color"]))
        edge_source["edge_size"].append(sum(edge_data["edge_size"]))
        edge_source["ind"].append(" ".join(edge_data["ind"]))
        edge_source["ind_size"].append(np.prod(edge_data["ind_size"]))
        edge_source["label"].append(" ".join(edge_data["label"]))

    node_source = collections.defaultdict(list)
    for _, node_data in G.nodes(data=True):
        if "ind" in node_data:
            continue

        x, y, *maybe_z = node_data["coo"]

        if "multi_colors" not in node_data:
            # single marker
            mcs = [node_data["color"]]
            mocs = [node_data["outline_color"]]
            szs = [node_data["size"]]
            os = node_data["outline_size"]
        elif multi_tag_style == "average":
            # plot a single marker with average color
            mcs = [average_color(node_data["multi_colors"])]
            mocs = [average_color(node_data["multi_outline_colors"])]
            szs = [node_data["size"]]
            os = node_data["outline_size"]
        elif multi_tag_style == "last":
            # plot a single marker with last tag
            mcs = [node_data["multi_colors"][-1]]
            mocs = [node_data["multi_outline_colors"][-1]]
            szs = [node_data["size"]]
            os = node_data["outline_size"]
        else:  # multi_tag_style in ("auto", "nest"):
            # plot multiple nested markers
            mcs = node_data["multi_colors"]
            mocs = node_data["multi_outline_colors"]
            szs = np.linspace(node_data["size"], 0, len(mcs) + 1)
            os = 0.0

        for mc, moc, sz in zip(mcs, mocs, szs):
            node_source["x"].append(x)
            node_source["y"].append(y)
            if maybe_z:
                node_source["z"].extend(maybe_z)

            node_source["color"].append(mc)
            node_source["outline_color"].append(moc)
            node_source["size"].append(sz)
            node_source["outline_size"].append(os)

            for k in ("hatch", "tags", "shape", "tid", "label"):
                node_source[k].append(node_data.get(k, None))

    return dict(edge_source), dict(node_source)


def _draw_matplotlib3d(G, **kwargs):
    import matplotlib.pyplot as plt

    edge_source, node_source = _linearize_graph_data(
        G, multi_tag_style=kwargs["multi_tag_style"]
    )

    ax = kwargs.pop("ax")
    if ax is None:
        fig = plt.figure(figsize=kwargs["figsize"])
        fig.patch.set_alpha(0.0)
        ax = plt.axes([0, 0, 1, 1], projection="3d")
        ax.patch.set_alpha(0.0)

        xmin = min(node_source["x"])
        xmax = max(node_source["x"])
        ymin = min(node_source["y"])
        ymax = max(node_source["y"])
        zmin = min(node_source["z"])
        zmax = max(node_source["z"])
        xyzmin = min((xmin, ymin, zmin))
        xyzmax = max((xmax, ymax, zmax))

        ax.set_xlim(xyzmin, xyzmax)
        ax.set_ylim(xyzmin, xyzmax)
        ax.set_zlim(xyzmin, xyzmax)
        ax.set_aspect("equal")
        ax.axis("off")

    # draw the edges
    # TODO: multiedges and left_inds
    for i in range(len(edge_source["x0"])):
        x0, x1 = edge_source["x0"][i], edge_source["x1"][i]
        xm = (x0 + x1) / 2
        y0, y1 = edge_source["y0"][i], edge_source["y1"][i]
        ym = (y0 + y1) / 2
        z0, z1 = edge_source["z0"][i], edge_source["z1"][i]
        zm = (z0 + z1) / 2
        ax.plot3D(
            [x0, x1],
            [y0, y1],
            [z0, z1],
            c=edge_source["color"][i],
            linewidth=edge_source["edge_size"][i],
        )
        label = edge_source["label"][i]
        if label:
            ax.text(
                xm,
                ym,
                zm,
                s=label,
                ha="center",
                va="center",
                color=edge_source["color"][i],
                fontsize=6,
            )

    node_source["color"] = [rgba[:3] for rgba in node_source["color"]]
    node_source["size"] = [100000 * s**2 for s in node_source["size"]]
    node_source["linewdith"] = [lw / 50 for lw in node_source["outline_size"]]

    # draw the nodes
    ax.scatter3D(
        xs="x",
        ys="y",
        zs="z",
        c="color",
        s="size",
        alpha=1.0,
        marker="o",
        data=node_source,
        depthshade=False,
        edgecolor=node_source["outline_color"],
        linewidth=node_source["outline_size"],
    )

    for _, node_data in G.nodes(data=True):
        label = node_data["label"]
        if label:
            ax.text(
                *node_data["coo"],
                s=label,
                ha="center",
                va="center",
                color=node_data["label_color"],
                fontsize=node_data["label_fontsize"],
                fontfamily=node_data["label_fontfamily"],
            )

    _add_legend_matplotlib(
        ax,
        kwargs["colors"],
        kwargs["legend"],
        kwargs["node_outline_darkness"],
        kwargs["label_color"],
        kwargs["font_family"],
    )

    if kwargs["return_fig"]:
        return fig
    else:
        plt.show()
        plt.close(fig)


def _draw_plotly(G, **kwargs):
    import plotly.graph_objects as go

    edge_source, node_source = _linearize_graph_data(
        G, multi_tag_style=kwargs["multi_tag_style"]
    )

    fig = go.Figure()
    fig.update_xaxes(visible=False)
    fig.update_yaxes(visible=False)
    fig.update_layout(
        width=100 * kwargs["figsize"][0],
        height=100 * kwargs["figsize"][1],
        margin=dict(l=10, r=10, b=10, t=10),
        plot_bgcolor="rgba(0,0,0,0)",
        paper_bgcolor="rgba(0,0,0,0)",
        showlegend=False,
        scene=dict(
            xaxis=dict(visible=False),
            yaxis=dict(visible=False),
            zaxis=dict(visible=False),
        ),
    )

    for i in range(len(edge_source["x0"])):
        x0, x1 = edge_source["x0"][i], edge_source["x1"][i]
        y0, y1 = edge_source["y0"][i], edge_source["y1"][i]
        xm, ym = (x0 + x1) / 2, (y0 + y1) / 2
        *rgb, alpha = edge_source["color"][i]
        edge_kwargs = dict(
            x=[x0, xm, x1],
            y=[y0, ym, y1],
            opacity=alpha,
            line=dict(
                color=to_rgba_str(rgb, 1.0),
                width=edge_source["edge_size"][i],
            ),
            customdata=[[edge_source["ind"][i], edge_source["ind_size"][i]]]
            * 2,
            # show ind and ind_size on hover:
            hovertemplate="%{customdata[0]}<br>size: %{customdata[1]}",
            mode="lines",
            name="",
        )
        if "z0" in edge_source:
            z0, z1 = edge_source["z0"][i], edge_source["z1"][i]
            zm = (z0 + z1) / 2
            edge_kwargs["z"] = [z0, zm, z1]
            # edges appear much thinner in 3D
            edge_kwargs["line"]["width"] *= 2
            fig.add_trace(go.Scatter3d(**edge_kwargs))
        else:
            fig.add_trace(go.Scatter(**edge_kwargs))

    node_kwargs = dict(
        x=node_source["x"],
        y=node_source["y"],
        marker=dict(
            opacity=1.0,
            color=list(map(to_rgba_str, node_source["color"])),
            size=[300 * s for s in node_source["size"]],
            line=dict(
                color=list(map(to_rgba_str, node_source["outline_color"])),
                width=2,
            ),
        ),
        customdata=list(
            zip(node_source["tid"], node_source["shape"], node_source["tags"])
        ),
        hovertemplate=(
            "tid: %{customdata[0]}<br>"
            "shape: %{customdata[1]}<br>"
            "tags: %{customdata[2]}"
        ),
        mode="markers",
        name="",
    )
    if "z" in node_source:
        node_kwargs["z"] = node_source["z"]
        fig.add_trace(go.Scatter3d(**node_kwargs))
    else:
        fig.add_trace(go.Scatter(**node_kwargs))
    fig.show()


# ---------------------------- layout functions ----------------------------- #


def _normalize_positions(pos):
    # normalize to unit square
    xmin = ymin = zmin = float("inf")
    xmax = ymax = zmax = float("-inf")
    for x, y, *maybe_z in pos.values():
        xmin = min(xmin, x)
        xmax = max(xmax, x)
        ymin = min(ymin, y)
        ymax = max(ymax, y)
        for z in maybe_z:
            zmin = min(zmin, z)
            zmax = max(zmax, z)

    # maintain aspect ratio:
    # center each dimension separately
    xmid, ymid, zmid = (xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2
    # but scale all dimensions by the largest range
    xdiameter, ydiameter, zdiameter = xmax - xmin, ymax - ymin, zmax - zmin
    radius = max((xdiameter, ydiameter, zdiameter)) / 2

    for node, (x, y, *maybe_z) in pos.items():
        pos[node] = (
            (x - xmid) / radius,
            (y - ymid) / radius,
            *((z - zmid) / radius for z in maybe_z),
        )

    return pos


def _rotate(xy, theta):
    """Return a rotated set of points."""
    s = np.sin(theta)
    c = np.cos(theta)

    xyr = np.empty_like(xy)
    xyr[:, 0] = c * xy[:, 0] - s * xy[:, 1]
    xyr[:, 1] = s * xy[:, 0] + c * xy[:, 1]

    return xyr


def _span(xy):
    """Return the vertical span of the points."""
    return xy[:, 1].max() - xy[:, 1].min()


def _massage_pos(pos, nangles=360, flatten=False):
    """Rotate a position dict's points to cover a small vertical span"""
    xy = np.empty((len(pos), 2))
    for i, (x, y) in enumerate(pos.values()):
        xy[i, 0] = x
        xy[i, 1] = y

    thetas = np.linspace(0, 2 * np.pi, nangles, endpoint=False)
    rxys = (_rotate(xy, theta) for theta in thetas)
    rxy0 = min(rxys, key=lambda rxy: _span(rxy))

    if flatten:
        rxy0[:, 1] /= 2

    return dict(zip(pos, rxy0))


def phyllotaxis_points(n):
    """J. Kogan, "A New Computationally Efficient Method for Spacing Points on
    a Sphere," Rose-Hulman Undergraduate Mathematics Journal, 18(2), 2017
    Article 5. scholar.rose-hulman.edu/rhumj/vol18/iss2/5.
    """

    def spherical_coordinate(x, y):
        return [np.cos(x) * np.cos(y), np.sin(x) * np.cos(y), np.sin(y)]

    x = 0.1 + 1.2 * n
    pts = []
    start = -1.0 + 1.0 / (n - 1.0)
    increment = (2.0 - 2.0 / (n - 1.0)) / (n - 1.0)
    for j in range(n):
        s = start + j * increment
        pts.append(
            spherical_coordinate(
                s * x,
                np.pi
                / 2.0
                * np.copysign(1, s)
                * (1.0 - np.sqrt(1.0 - abs(s))),
            )
        )
    return pts


def layout_single_tensor(tn, dim=2):
    """Manually layout indices around a tensor either in a circle or sphere."""
    ((tid, t),) = tn.tensor_map.items()

    if dim == 2.5:
        dim = 3
        project_back_to_2d = True
    else:
        project_back_to_2d = False

    pos = {tid: (0.0,) * dim}
    if dim == 2:
        # fix around a circle
        angles = np.linspace(0, 2 * np.pi, t.ndim, endpoint=False)
        for ind, angle in zip(t.inds, angles):
            pos[ind] = (-np.cos(angle), np.sin(angle))
    else:
        # fix around a sphere
        for ind, coo in zip(t.inds, phyllotaxis_points(t.ndim)):
            pos[ind] = coo

    if project_back_to_2d:
        pos = {k: v[:2] for k, v in pos.items()}

    return pos


def layout_networkx(
    G,
    layout="kamada_kawai",
    pos0=None,
    fixed=None,
    dim=2,
    **kwargs,
):
    import networkx as nx

    layout_fn = getattr(nx, layout + "_layout")

    if pos0 is not None:
        if layout not in ("spring", "kamada_kawai"):
            warnings.warn(
                "Initial positions supplied but layout is not spring-based, "
                "so `pos0` is being ignored."
            )
        else:
            kwargs["pos"] = pos0

    if fixed is not None:
        if layout != "spring":
            warnings.warn(
                "Fixed positions supplied but layout is not spring-based, "
                "so `fixed` is being ignored."
            )
        else:
            kwargs["fixed"] = fixed

    return layout_fn(G, dim=dim, **kwargs)


def layout_pygraphviz(
    G,
    layout="neato",
    pos0=None,
    fixed=None,
    dim=2,
    iterations=None,
    k=None,
    **kwargs,
):
    # TODO: max iters
    # TODO: spring parameter
    # TODO: work out why pos0 and fix don't work
    import pygraphviz as pgv

    if k is not None:
        warnings.warn(
            "`k` is being ignored as layout is being done by pygraphviz."
        )

    aG = pgv.AGraph()

    # create nodes
    if pos0 is not None:
        fixed = fixed or set()
        for node, coo in pos0.items():
            pos = ",".join((f"{w:f}" for w in coo))
            pin = "true" if node in fixed else "false"
            aG.add_node(str(node), pos=pos, pin=pin)

        warnings.warn(
            "Initial and fixed positions don't seem "
            "to work currently with pygraphviz."
        )

    # create edges
    mapping = {}
    for nodea, nodeb in G.edges():
        s_nodea = str(nodea)
        s_nodeb = str(nodeb)
        mapping[s_nodea] = nodea
        mapping[s_nodeb] = nodeb
        aG.add_edge(s_nodea, s_nodeb)

    # layout options
    if iterations is not None:
        kwargs["maxiter"] = iterations
    if dim == 2.5:
        kwargs["dim"] = 3
        kwargs["dimen"] = 2
    else:
        kwargs["dim"] = kwargs["dimen"] = dim
    args = " ".join(f"-G{k}={v}" for k, v in kwargs.items())

    # run layout algorithm
    aG.layout(prog=layout, args=args)

    # extract layout
    pos = {}
    for snode, node in mapping.items():
        spos = aG.get_node(snode).attr["pos"]
        pos[node] = tuple(map(float, spos.split(",")))

    pos = _normalize_positions(pos)
    if dim < 3:
        pos = _massage_pos(pos)

    return pos


def get_positions(
    tn,
    G,
    *,
    dim=2,
    fix=None,
    layout="auto",
    initial_layout="auto",
    refine_layout="auto",
    iterations="auto",
    k=None,
):
    if (tn is not None) and (tn.num_tensors == 1) and (fix is None):
        # single tensor, layout manually
        return layout_single_tensor(tn, dim=dim)

    if layout != "auto":
        # don't use two step layout with relaxation
        initial_layout = layout
        iterations = 0

    if fix is None:
        fix = dict()
    else:
        fix = parse_dict_to_tids_or_inds(fix, tn)
        # find range with which to scale spectral points with
        xmin, xmax, ymin, ymax = (
            f(fix.values(), key=lambda xy: xy[i])[i]
            for f, i in [(min, 0), (max, 0), (min, 1), (max, 1)]
        )
        if xmin == xmax:
            xmin, xmax = xmin - 1, xmax + 1
        if ymin == ymax:
            ymin, ymax = ymin - 1, ymax + 1
        xymin, xymax = min(xmin, ymin), max(xmax, ymax)

    if all(node in fix for node in G.nodes):
        # everything is already fixed -> simply normalize
        return _normalize_positions(fix)

    if initial_layout == "auto":
        # automatically select
        if len(G) <= 100:
            # usually nicest
            initial_layout = "kamada_kawai"
        else:
            # faster, but not as nice
            initial_layout = "spectral"

    if refine_layout == "auto":
        # automatically select
        refine_layout = "spring"
        # if len(G) <= 100:
        #     # usually nicest
        #     refine_layout = "fdp"
        # else:
        #     # faster, but not as nice
        #     refine_layout = "sfdp"

    if iterations == "auto":
        # the smaller the graph, the more iterations we can afford
        iterations = max(200, 1000 - len(G))

    if dim == 2.5:
        dim = 3
        project_back_to_2d = True
    else:
        project_back_to_2d = False

    # use spectral or other layout as starting point
    if initial_layout in ("neato", "fdp", "sfdp", "dot"):
        pos0 = layout_pygraphviz(G, initial_layout, dim=dim)
    else:
        pos0 = layout_networkx(G, initial_layout, dim=dim)

    # scale points to fit with specified positions
    if fix:
        # but update with fixed positions
        pos0.update(
            valmap(
                lambda xy: np.array(
                    (
                        2 * (xy[0] - xymin) / (xymax - xymin) - 1,
                        2 * (xy[1] - xymin) / (xymax - xymin) - 1,
                    )
                ),
                fix,
            )
        )
        fixed = fix.keys()
    else:
        fixed = None

    # and then relax remaining using spring layout
    if iterations:
        if refine_layout == "spring":
            pos = layout_networkx(
                G,
                "spring",
                pos0=pos0,
                fixed=fixed,
                k=k,
                dim=dim,
                iterations=iterations,
            )
        elif refine_layout in ("fdp", "sfdp", "neato"):
            # XXX: currently doesn't seem to work with pos0 and fixed
            pos = layout_pygraphviz(
                G,
                refine_layout,
                pos0=pos0,
                fixed=fixed,
                k=k,
                dim=dim,
                iterations=iterations,
            )
        else:
            raise ValueError(f"Unknown refining layout {refine_layout}.")
    else:
        # no relaxation
        pos = pos0

    if project_back_to_2d:
        # ignore z-coordinate
        pos = {k: v[:2] for k, v in pos.items()}
        dim = 2

    # map all to range [-1, +1], but preserving aspect ratio
    pos = _normalize_positions(pos)

    if (not fix) and (dim == 2):
        # finally rotate them to cover a small vertical span
        pos = _massage_pos(pos)

    return pos


# ----------------------------- color functions ----------------------------- #


def get_colors(color, custom_colors=None, alpha=None):
    """Generate a sequence of rgbs for tag(s) ``color``."""
    from matplotlib.colors import to_rgba

    from ..schematic import auto_colors

    if color is None:
        return dict()

    if isinstance(color, str):
        color = (color,)

    if custom_colors is not None:
        rgbs = [to_rgba(c, alpha=alpha) for c in custom_colors]
        return dict(zip(color, rgbs))

    nc = len(color)
    if nc <= 7:
        rgbs = auto_colors(nc, alpha=alpha, default_sequence=True)
        return dict(zip(color, rgbs))

    rgbs = auto_colors(nc, alpha)
    return dict(zip(color, rgbs))


def to_rgba_str(color, alpha=None):
    from matplotlib.colors import to_rgba

    rgba = to_rgba(color, alpha)
    r = int(rgba[0] * 255) if isinstance(rgba[0], float) else rgba[0]
    g = int(rgba[1] * 255) if isinstance(rgba[1], float) else rgba[1]
    b = int(rgba[2] * 255) if isinstance(rgba[2], float) else rgba[2]
    return f"rgba({r}, {g}, {b}, {rgba[3]})"


def auto_color_html(s):
    """Automatically hash and color a string for HTML display."""
    from ..schematic import hash_to_color

    if not isinstance(s, str):
        s = str(s)
    return f'<b style="color: {hash_to_color(s)};">{s}</b>'


# ---------------------------- tensor functions ----------------------------- #


def visualize_tensor(tensor, **kwargs):
    """Visualize all entries of a tensor, with indices mapped into the plane
    and values mapped into a color wheel.

    Parameters
    ----------
    tensor : Tensor
        The tensor to visualize.
    skew_factor : float, optional
        When there are more than two dimensions, a factor to scale the
        rotations by to avoid overlapping data points.
    size_map : bool, optional
        Whether to map the tensor value magnitudes to marker size.
    size_scale : float, optional
        An overall factor to scale the marker size by.
    alpha_map : bool, optional
        Whether to map the tensor value magnitudes to marker alpha.
    alpha_pow : float, optional
        The power to raise the magnitude to when mapping to alpha.
    alpha : float, optional
        The overall alpha to use for all markers if ``not alpha_map``.
    show_lattice : bool, optional
        Show a small grey dot for every 'lattice' point regardless of value.
    lattice_opts : dict, optional
        Options to pass to ``maplotlib.Axis.scatter`` for the lattice points.
    linewidths : float, optional
        The linewidth to use for the markers.
    marker : str, optional
        The marker to use for the markers.
    figsize : tuple, optional
        The size of the figure to create, if ``ax`` is not provided.
    ax : matplotlib.Axis, optional
        The axis to draw to. If not provided, a new figure will be created.

    Returns
    -------
    fig : matplotlib.Figure
        The figure containing the plot, or ``None`` if ``ax`` was provided.
    ax : matplotlib.Axis
        The axis containing the plot.
    """
    import xyzpy as xyz

    kwargs.setdefault("legend", True)
    kwargs.setdefault("compass", True)
    kwargs.setdefault("compass_labels", tensor.inds)
    return xyz.visualize_tensor(tensor.data, **kwargs)


def choose_squarest_grid(x):
    p = x**0.5
    if p.is_integer():
        m = n = int(p)
    else:
        m = int(round(p))
        p = int(p)
        n = p if m * p >= x else p + 1
    return m, n


def visualize_tensors(
    tn,
    mode="network",
    r=None,
    r_scale=1.0,
    figsize=None,
    **visualize_opts,
):
    """Visualize all the entries of every tensor in this network.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to visualize.
    mode : {'network', 'grid', 'row', 'col'}, optional
        How to arrange each tensor's visualization.

        - ``'network'``: arrange each tensor's visualization according to the
            automatic layout given by ``draw``.
        - ``'grid'``: arrange each tensor's visualization in a grid.
        - ``'row'``: arrange each tensor's visualization horizontally.
        - ``'col'``: arrange each tensor's visualization vertically.

    r : float, optional
        The absolute radius of each tensor's visualization, when
        ``mode='network'``.
    r_scale : float, optional
        A relative scaling factor for the radius of each tensor's
        visualization, when ``mode='network'``.
    figsize : tuple, optional
        The size of the figure to create, if ``ax`` is not provided.
    visualize_opts
        Supplied to ``visualize_tensor``.
    """
    from matplotlib import pyplot as plt

    if figsize is None:
        figsize = (2 * tn.num_tensors**0.4, 2 * tn.num_tensors**0.4)
    if r is None:
        r = 1.0 / tn.num_tensors**0.5
    r *= r_scale

    max_mag = None
    visualize_opts.setdefault("max_mag", max_mag)
    visualize_opts.setdefault("size_scale", r)

    if mode == "network":
        fig = plt.figure(figsize=figsize)
        pos = tn.draw(get="pos")
        for tid, (x, y) in pos.items():
            if tid not in tn.tensor_map:
                # hyper indez
                continue
            x = (x + 1) / 2 - r / 2
            y = (y + 1) / 2 - r / 2
            ax = fig.add_axes((x, y, r / 2, r / 2))
            tn.tensor_map[tid].visualize(ax=ax, **visualize_opts)
    else:
        if mode == "grid":
            px, py = choose_squarest_grid(tn.num_tensors)
        elif mode == "row":
            px, py = tn.num_tensors, 1
            figsize = (2 * figsize[0], figsize[1] / 2)
        elif mode == "col":
            px, py = 1, tn.num_tensors
            figsize = (figsize[0] / 2, 2 * figsize[1])

        fig, axs = plt.subplots(py, px, figsize=figsize)
        for i, t in enumerate(tn):
            t.visualize(ax=axs.flat[i], **visualize_opts)
        for ax in axs.flat[i:]:
            ax.set_axis_off()

    # transparent background
    fig.patch.set_alpha(0.0)

    plt.show()
    plt.close()



import itertools

import pytest
import autoray as ar
import numpy as np
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


class TestPEPSConstruct:
    @pytest.mark.parametrize("Lx", [3, 4, 5])
    @pytest.mark.parametrize("Ly", [3, 4, 5])
    def test_basic_rand(self, Lx, Ly):
        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=4)

        assert psi.max_bond() == 4
        assert psi.Lx == Lx
        assert psi.Ly == Ly
        assert len(psi.tensor_map) == Lx * Ly
        assert psi.site_inds == tuple(
            f"k{i},{j}" for i in range(Lx) for j in range(Ly)
        )
        assert psi.site_tags == tuple(
            f"I{i},{j}" for i in range(Lx) for j in range(Ly)
        )

        assert psi.bond_size((1, 1), (1, 2)) == (4)

        for i in range(Lx):
            assert len(psi.select(f"X{i}").tensor_map) == Ly
        for j in range(Ly):
            assert len(psi.select(f"Y{j}").tensor_map) == Lx

        for i in range(Lx):
            for j in range(Ly):
                assert psi.phys_dim(i, j) == 2
                assert isinstance(psi[i, j], qtn.Tensor)
                assert isinstance(psi[f"I{i},{j}"], qtn.Tensor)

        if Lx == Ly == 3:
            psi_dense = psi.to_qarray(optimize="auto-hq")
            assert psi_dense.shape == (512, 1)

        psi.show()
        assert f"Lx={Lx}" in psi.__str__()
        assert f"Lx={Lx}" in psi.__repr__()

    def test_cyclic_edge_cases(self):
        peps = qtn.PEPS.rand(3, 3, bond_dim=1, cyclic=True)
        assert peps.is_cyclic_x()
        assert peps.is_cyclic_y()
        assert peps.num_indices == peps.num_tensors * 3

    def test_zeros(self):
        peps = qtn.PEPS.zeros(3, 3, cyclic=True, bond_dim=1)
        assert peps.num_tensors == 9
        assert peps.num_indices == 27
        assert_allclose(peps.to_dense(), np.zeros([512, 1]))

    def test_flatten(self):
        psi = qtn.PEPS.rand(3, 5, 3, seed=42)
        norm = psi.H & psi
        assert len(norm.tensors) == 30
        norm.flatten_()
        assert len(norm.tensors) == 15
        assert norm.max_bond() == 9

    def test_add_peps(self):
        pa = qtn.PEPS.rand(3, 4, 2)
        pb = qtn.PEPS.rand(3, 4, 3)
        pc = qtn.PEPS.rand(3, 4, 4)
        pab = pa + pb
        assert pab.max_bond() == 5
        assert pab @ pc == pytest.approx(pa @ pc + pb @ pc)

    @pytest.mark.parametrize("Lx", [3, 4, 5])
    @pytest.mark.parametrize("Ly", [3, 4, 5])
    def test_bond_coordinates(self, Lx, Ly):
        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=1)
        all_bonds = tuple(psi.gen_bond_coos())
        assert len(all_bonds) == 2 * Lx * Ly - Lx - Ly
        he = tuple(psi.gen_horizontal_even_bond_coos())
        ho = tuple(psi.gen_horizontal_odd_bond_coos())
        ve = tuple(psi.gen_vertical_even_bond_coos())
        vo = tuple(psi.gen_vertical_odd_bond_coos())
        for p in (he, ho, ve, vo):
            assert len(set(p)) == len(p)
            # check there is no overlap at all
            sites = tuple(itertools.chain.from_iterable(he))
            assert len(set(sites)) == len(sites)
        # check all coordinates are generated
        assert set(itertools.chain(he, ho, ve, vo)) == set(all_bonds)

    @pytest.mark.parametrize(
        "where",
        [
            [(0, 0)],
            [(0, 1)],
            [(0, 2)],
            [(2, 2)],
            [(3, 2)],
            [(3, 1)],
            [(3, 0)],
            [(2, 0)],
            [(1, 1)],
        ],
    )
    @pytest.mark.parametrize("contract", [False, True])
    def test_gate_2d_single_site(self, where, contract):
        Lx = 4
        Ly = 3
        D = 2

        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=D, seed=42, dtype=complex)
        psi_d = psi.to_qarray()
        G = qu.rand_matrix(2)

        # compute the exact dense reference
        dims = [[2] * Ly] * Lx
        IGI = qu.ikron(G, dims, where, sparse=True)
        xe = (psi_d.H @ IGI @ psi_d).item()

        tn = psi.H & psi.gate(G, where, contract=contract)
        assert len(tn.tensors) == 2 * Lx * Ly + int(not contract)

        assert tn ^ all == pytest.approx(xe)

    @pytest.mark.parametrize(
        "contract", [False, True, "split", "reduce-split"]
    )
    @pytest.mark.parametrize(
        "where",
        [
            [(1, 1), (2, 1)],
            [(3, 2), (2, 2)],
            # XXX: long range, disable for now
            # [(0, 0), (1, 1)],
            # [(3, 1), (1, 2)],
        ],
    )
    def test_gate_2d_two_site(self, where, contract):
        Lx = 4
        Ly = 3
        D = 2

        psi = qtn.PEPS.rand(Lx, Ly, bond_dim=D, seed=42, dtype=complex)
        psi_d = psi.to_qarray()

        # ikron can't tensor operators across non-adjacent subsytems
        # so we explicitly construct the gate as a sum of tensor components
        G_comps = [(qu.rand_matrix(2), qu.rand_matrix(2)) for _ in range(4)]
        G = sum(A & B for A, B in G_comps)

        # compute the exact dense reference
        dims = [[2] * Ly] * Lx
        IGI = sum(
            qu.ikron([A, B], dims, where, sparse=True) for A, B in G_comps
        )

        xe = (psi_d.H @ IGI @ psi_d).item()

        tn = psi.H & psi.gate(G, where, contract=contract)
        change = {False: 1, True: -1, "split": 0, "reduce-split": 0}[contract]
        assert len(tn.tensors) == 2 * Lx * Ly + change

        assert tn ^ all == pytest.approx(xe)

    @pytest.mark.parametrize(
        "propagate_tags", [False, True, "sites", "register"]
    )
    def test_gate_propagate_tags(self, propagate_tags):
        Lx = 4
        Ly = 3
        D = 1
        psi = qtn.PEPS.rand(Lx, Ly, D, tags="PSI0")
        psi.gate_(
            qu.rand_uni(4),
            [(1, 1), (1, 2)],
            tags="G1",
            propagate_tags=propagate_tags,
        )
        psi.gate_(
            qu.rand_uni(4),
            [(1, 2), (3, 2)],
            tags="G2",
            propagate_tags=propagate_tags,
        )
        if propagate_tags is False:
            assert set(psi["G1"].tags) == {"G1"}
            assert set(psi["G2"].tags) == {"G2"}
        if propagate_tags is True:
            tgs1 = {"I1,1", "I1,2", "G1", "PSI0", "Y1", "Y2", "X1"}
            assert set(psi["G1"][0].tags) == tgs1
            assert set(psi["G2"].tags) == tgs1 | {"G2", "I3,2", "X3", "Y2"}
        if propagate_tags == "sites":
            assert set(psi["G1"].tags) == {"G1", "I1,1", "I1,2"}
            assert set(psi["G2"].tags) == {"G2", "I1,1", "I1,2", "I3,2"}
        if propagate_tags == "register":
            assert set(psi["G1"].tags) == {"G1", "I1,1", "I1,2"}
            assert set(psi["G2"].tags) == {"G2", "I1,2", "I3,2"}


class Test2DContract:
    @pytest.mark.parametrize("mode", ["mps", "projector", "full-bond"])
    def test_contract_boundary(self, mode):
        # make a large but cheap and easy (mostly positive) TN
        rng = np.random.default_rng(42)
        tn = qtn.TN2D_from_fill_fn(
            lambda shape: rng.uniform(low=-0.1, size=shape),
            Lx=8,
            Ly=8,
            D=2,
        )
        Zex = tn.contract(...)
        Z = tn.contract_boundary(max_bond=4, mode=mode)
        assert Z == pytest.approx(Zex, rel=1e-3)

    def test_contract_2d_one_layer_boundary(self):
        psi = qtn.PEPS.rand(4, 4, 3, seed=42)
        norm = psi.make_norm()
        xe = norm.contract(all, optimize="auto-hq")
        xt = norm.contract_boundary(max_bond=9)
        assert xt == pytest.approx(xe, rel=1e-2)

    def test_contract_2d_two_layer_boundary(self):
        psi = qtn.PEPS.rand(4, 4, 3, seed=42, tags="KET")
        norm = psi.make_norm()
        xe = norm.contract(all, optimize="auto-hq")
        xt = norm.contract_boundary(max_bond=27, layer_tags=["KET", "BRA"])
        assert xt == pytest.approx(xe, rel=1e-2)

    def test_contract_2d_full_bond(self):
        psi = qtn.PEPS.rand(4, 4, 3, seed=42, tags="KET")
        norm = psi.make_norm()
        xe = norm.contract(all, optimize="auto-hq")
        xt = norm.contract_boundary(max_bond=27, mode="full-bond")
        assert xt == pytest.approx(xe, rel=1e-2)

    @pytest.mark.parametrize("dims", [(10, 4), (4, 10)])
    def test_contract_boundary_stopping_criterion(self, dims):
        tn = qtn.TN2D_from_fill_fn(
            lambda shape: ar.lazy.Variable(shape=shape, backend="numpy"),
            *dims,
            D=2,
        )
        tn.contract_ctmrg_(4, cutoff=0.0, final_contract=False, progbar=True)
        assert tn.max_bond() == 4
        assert 16 <= tn.num_tensors <= 20

    @pytest.mark.parametrize("lazy", [False, True])
    def test_coarse_grain_basics(self, lazy):
        tn = qtn.TN2D_from_fill_fn(
            lambda shape: ar.lazy.Variable(shape, backend="numpy"),
            Lx=6,
            Ly=7,
            D=2,
        )
        tncg = tn.coarse_grain_hotrg("x", max_bond=3, cutoff=0.0, lazy=lazy)
        assert (tncg.Lx, tncg.Ly) == (3, 7)
        assert not tncg.outer_inds()
        assert tncg.max_bond() == 3
        assert "I4,0" not in tncg.tag_map
        assert "X5" not in tncg.tag_map

        tncg = tn.coarse_grain_hotrg("y", max_bond=3, cutoff=0.0, lazy=lazy)
        assert (tncg.Lx, tncg.Ly) == (6, 4)
        assert not tncg.outer_inds()
        assert tncg.max_bond() == 3
        assert "I0,5" not in tncg.tag_map
        assert "Y6" not in tncg.tag_map

    def test_contract_hotrg(self):
        tn = qtn.TN2D_classical_ising_partition_function(16, 16, 0.44)
        tn.contract_hotrg_(max_bond=5, progbar=True, equalize_norms=1.0)
        Zap = tn.item() * 10**tn.exponent
        assert Zap == pytest.approx(8.459419593253275e100, rel=2e-3)

    def test_contract_hotrg_two_layer_rand_peps(self):
        rng = np.random.default_rng(42)
        psi = qtn.PEPS.from_fill_fn(
            lambda shape: rng.uniform(low=-0.1, size=shape),
            Lx=7,
            Ly=5,
            bond_dim=2,
        )
        norm = psi.make_norm()
        xe = norm.contract(all, optimize="auto-hq")
        xt = norm.contract_hotrg(max_bond=5)
        assert xt == pytest.approx(xe, rel=1e-4)

    def test_ising_accuracy_regression(self):
        tn = qtn.TN2D_classical_ising_partition_function(16, 16, 0.44)
        for s in [("xmin",), ("xmax",), ("ymin",), ("ymax",)]:
            Zap = tn.contract_boundary(max_bond=8, sequence=s)
            assert Zap == pytest.approx(8.459419593253275e100, rel=2.2e-7)
        for s in [("xmin", "xmax"), ("ymin", "ymax")]:
            Zap = tn.contract_boundary(max_bond=8, sequence=s)
            assert Zap == pytest.approx(8.459419593253275e100, rel=3.9e-9)

    @pytest.mark.parametrize("mode", ["mps", "ctmrg", "hotrg"])
    def test_cdl_rand_large(self, mode):
        tn = qtn.TN2D_rand_hidden_loop(10, 10, seed=42, contract_sites=False)
        Zex = tn.contract(...)
        tn = qtn.TN2D_rand_hidden_loop(10, 10, seed=42, contract_sites=True)

        if mode == "mps":
            Z = tn.contract_boundary(max_bond=16)
        elif mode == "ctmrg":
            Z = tn.contract_ctmrg(max_bond=16)
        elif mode == "hotrg":
            Z = tn.contract_hotrg(max_bond=16)

        assert Z == pytest.approx(Zex, rel=1e-1)

    @pytest.mark.parametrize(
        "mode,two_layer",
        [
            ("mps", False),
            ("mps", True),
            ("full-bond", False),
        ],
    )
    def test_compute_x_envs(self, mode, two_layer):
        psi = qtn.PEPS.rand(5, 4, 2, seed=42, tags="KET")
        norm = psi.make_norm()
        ex = norm.contract(all)

        if two_layer:
            compress_opts = {
                "cutoff": 1e-6,
                "max_bond": 12,
                "mode": mode,
                "layer_tags": ["KET", "BRA"],
            }
        else:
            compress_opts = {"cutoff": 1e-6, "max_bond": 8, "mode": mode}
        row_envs = norm.compute_x_environments(**compress_opts)

        for i in range(norm.Lx):
            norm_i = (
                row_envs["xmin", i]
                & norm.select(norm.x_tag(i))
                & row_envs["xmax", i]
            )
            x = norm_i.contract(all)
            assert x == pytest.approx(ex, rel=1e-2)

    @pytest.mark.parametrize(
        "mode,two_layer",
        [
            ("mps", False),
            ("mps", True),
            ("full-bond", False),
        ],
    )
    def test_compute_y_envs(self, mode, two_layer):
        psi = qtn.PEPS.rand(4, 5, 2, seed=42, tags="KET")
        norm = psi.retag({"KET": "BRA"}).H | psi
        ex = norm.contract(all)

        if two_layer:
            compress_opts = {
                "cutoff": 1e-6,
                "max_bond": 12,
                "mode": mode,
                "layer_tags": ["KET", "BRA"],
            }
        else:
            compress_opts = {"cutoff": 1e-6, "max_bond": 8, "mode": mode}
        col_envs = norm.compute_y_environments(**compress_opts)

        for j in range(norm.Lx):
            norm_j = (
                col_envs["ymin", j]
                & norm.select(norm.y_tag(j))
                & col_envs["ymax", j]
            )
            x = norm_j.contract(all)
            assert x == pytest.approx(ex, rel=1e-2)

    def test_normalize(self):
        psi = qtn.PEPS.rand(4, 5, 2, seed=42)
        norm = (psi.H | psi).contract(all)
        assert norm != pytest.approx(1.0)
        psi.normalize_(balance_bonds=True, equalize_norms=True, cutoff=2e-3)
        norm = (psi.H | psi).contract(all)
        assert norm == pytest.approx(1.0, rel=0.01)

    @pytest.mark.parametrize("normalized", [False, True])
    @pytest.mark.parametrize("mode", ["mps", "full-bond"])
    def test_compute_local_expectation_one_sites(self, mode, normalized):
        peps = qtn.PEPS.rand(4, 3, 2, seed=42, dtype="complex")

        # reference
        k = peps.to_qarray()
        if normalized:
            qu.normalize(k)
        coos = list(itertools.product([0, 2, 3], [0, 1, 2]))
        terms = {coo: qu.rand_matrix(2) for coo in coos}
        dims = [[2] * 3] * 4
        A = sum(
            qu.ikron(A, dims, [coo], sparse=True) for coo, A in terms.items()
        )
        ex = qu.expec(A, k)

        opts = dict(cutoff=2e-3, max_bond=9, contract_optimize="auto-hq")
        e = peps.compute_local_expectation(
            terms, mode=mode, normalized=normalized, **opts
        )

        assert e == pytest.approx(ex, rel=1e-2)

    @pytest.mark.parametrize("normalized", [False, True])
    @pytest.mark.parametrize("mode", ["mps", "full-bond"])
    def test_compute_local_expectation_two_sites(self, mode, normalized):
        H = qu.ham_heis_2D(4, 3, sparse=True)
        Hij = qu.ham_heis(2, cyclic=False)

        peps = qtn.PEPS.rand(4, 3, 2, seed=42)
        k = peps.to_qarray()

        if normalized:
            qu.normalize(k)
        ex = qu.expec(H, k)

        opts = dict(
            mode=mode,
            normalized=normalized,
            cutoff=2e-3,
            max_bond=16,
            contract_optimize="auto-hq",
        )

        # compute 2x1 and 1x2 plaquettes separately
        hterms = {coos: Hij for coos in peps.gen_horizontal_bond_coos()}
        vterms = {coos: Hij for coos in peps.gen_vertical_bond_coos()}

        he = peps.compute_local_expectation(hterms, **opts)
        ve = peps.compute_local_expectation(vterms, **opts)

        assert he + ve == pytest.approx(ex, rel=1e-2)

        # compute all terms in 2x2 plaquettes
        terms_all = {**hterms, **vterms}
        e = peps.compute_local_expectation(terms_all, autogroup=False, **opts)

        assert e == pytest.approx(ex, rel=1e-2)

    def test_cyclic_basic(self):
        tn = qtn.TN2D_rand(Lx=3, Ly=4, D=2, cyclic=True)
        assert tn.is_cyclic_x()
        assert tn.is_cyclic_y()
        assert tn.num_indices == 2 * 3 * 4
        tn = qtn.TN2D_rand(Lx=3, Ly=4, D=2, cyclic=(False, True))
        assert not tn.is_cyclic_x()
        assert tn.is_cyclic_y()
        assert tn.num_indices == 2 * 3 * 4 - 4
        tn = qtn.TN2D_rand(Lx=3, Ly=4, D=2, cyclic=(True, False))
        assert tn.is_cyclic_x()
        assert not tn.is_cyclic_y()
        assert tn.num_indices == 2 * 3 * 4 - 3
        tn = qtn.TN2D_rand(Lx=3, Ly=4, D=2, cyclic=(False, False))
        assert not tn.is_cyclic_x()
        assert not tn.is_cyclic_y()
        assert tn.num_indices == 2 * 3 * 4 - 7

    @pytest.mark.parametrize("cyclicx", [False, True])
    @pytest.mark.parametrize("cyclicy", [False, True])
    @pytest.mark.parametrize("mode", ["mps", "hotrg", "ctmrg"])
    def test_cyclic_contract(self, cyclicx, cyclicy, mode):
        Lx = 5
        Ly = 6
        D = 2
        chi = 3
        tn = qtn.TN2D_rand(
            Lx,
            Ly,
            D,
            cyclic=(cyclicx, cyclicy),
            seed=42,
            dist="uniform",
        )
        Zex = tn.contract(...)
        if mode == "hotrg":
            Z = tn.contract_hotrg(chi)
        elif mode == "ctmrg":
            Z = tn.contract_ctmrg(chi)
        else:
            Z = tn.contract_boundary(chi, mode=mode)
        assert abs(1 - Z / Zex) < 1e-3


class TestPEPO:
    @pytest.mark.parametrize("Lx", [3, 4, 5])
    @pytest.mark.parametrize("Ly", [3, 4, 5])
    def test_basic_rand(self, Lx, Ly):
        X = qtn.PEPO.rand_herm(Lx, Ly, bond_dim=4)

        assert X.max_bond() == 4
        assert X.Lx == Lx
        assert X.Ly == Ly
        assert len(X.tensor_map) == Lx * Ly
        assert X.upper_inds == tuple(
            f"k{i},{j}" for i in range(Lx) for j in range(Ly)
        )
        assert X.lower_inds == tuple(
            f"b{i},{j}" for i in range(Lx) for j in range(Ly)
        )
        assert X.site_tags == tuple(
            f"I{i},{j}" for i in range(Lx) for j in range(Ly)
        )

        assert X.bond_size((1, 1), (1, 2)) == (4)

        for i in range(Lx):
            assert len(X.select(f"X{i}").tensor_map) == Ly
        for j in range(Ly):
            assert len(X.select(f"Y{j}").tensor_map) == Lx

        for i in range(Lx):
            for j in range(Ly):
                assert X.phys_dim(i, j) == 2
                assert isinstance(X[i, j], qtn.Tensor)
                assert isinstance(X[f"I{i},{j}"], qtn.Tensor)

        if Lx == Ly == 3:
            X_dense = X.to_qarray(optimize="auto-hq")
            assert X_dense.shape == (512, 512)
            assert qu.isherm(X_dense)

        X.show()
        assert f"Lx={Lx}" in X.__str__()
        assert f"Lx={Lx}" in X.__repr__()

    def test_add_pepo(self):
        pa = qtn.PEPO.rand(3, 4, 2)
        pb = qtn.PEPO.rand(3, 4, 3)
        pc = qtn.PEPO.rand(3, 4, 4)
        pab = pa + pb
        assert pab.max_bond() == 5
        assert pab @ pc == pytest.approx(pa @ pc + pb @ pc)

    def test_apply_pepo(self):
        A = qtn.PEPO.rand(Lx=3, Ly=2, bond_dim=2, seed=1)
        x = qtn.PEPS.rand(Lx=3, Ly=2, bond_dim=2, seed=0)
        y = A.apply(x)
        assert y.num_indices == x.num_indices
        Ad = A.to_qarray()
        xd = x.to_qarray()
        yd = y.to_qarray()
        assert_allclose(Ad @ xd, yd)
        yc = A.apply(x, compress=True, max_bond=3)
        assert yc.max_bond() == 3


class TestMisc:
    def test_calc_plaquette_sizes(self):
        from quimb.tensor.tensor_2d import calc_plaquette_sizes

        H2 = {None: qu.ham_heis(2)}
        ham = qtn.LocalHam2D(10, 10, H2)
        assert calc_plaquette_sizes(ham.terms.keys()) == ((1, 2), (2, 1))
        assert calc_plaquette_sizes(ham.terms.keys(), autogroup=False) == (
            (2, 2),
        )
        H2[(1, 1), (2, 2)] = 0.5 * qu.ham_heis(2)
        ham = qtn.LocalHam2D(10, 10, H2)
        assert calc_plaquette_sizes(ham.terms.keys()) == ((2, 2),)
        H2[(2, 2), (2, 4)] = 0.25 * qu.ham_heis(2)
        H2[(2, 4), (4, 4)] = 0.25 * qu.ham_heis(2)
        ham = qtn.LocalHam2D(10, 10, H2)
        assert calc_plaquette_sizes(ham.terms.keys()) == (
            (1, 3),
            (2, 2),
            (3, 1),
        )
        assert calc_plaquette_sizes(ham.terms.keys(), autogroup=False) == (
            (3, 3),
        )

    def test_calc_plaquette_map(self):
        from quimb.tensor.tensor_2d import calc_plaquette_map

        plaquettes = [
            # 2x2 plaquette covering all sites
            ((0, 0), (2, 2)),
            # horizontal plaquettes
            ((0, 0), (1, 2)),
            ((1, 0), (1, 2)),
            # vertical plaquettes
            ((0, 0), (2, 1)),
            ((0, 1), (2, 1)),
        ]
        assert calc_plaquette_map(plaquettes) == {
            (0, 0): ((0, 0), (2, 1)),
            (0, 1): ((0, 1), (2, 1)),
            (1, 0): ((1, 0), (1, 2)),
            (1, 1): ((1, 0), (1, 2)),
            ((0, 0), (0, 1)): ((0, 0), (1, 2)),
            ((0, 0), (1, 0)): ((0, 0), (2, 1)),
            ((0, 0), (1, 1)): ((0, 0), (2, 2)),
            ((0, 1), (1, 0)): ((0, 0), (2, 2)),
            ((0, 1), (1, 1)): ((0, 1), (2, 1)),
            ((1, 0), (1, 1)): ((1, 0), (1, 2)),
        }



"""Generic methods for compressing arbitrary geometry tensor networks, where
the tensor network can locally have arbitrary structure and outer indices.

- [x] projector
- [x] l2bp
- [x] local early
- [x] local late
- [x] superorthogonal

"""

from ..utils import ensure_dict
from .tensor_arbgeom import create_lazy_edge_map
from .tensor_core import TensorNetwork, choose_local_compress_gauge_settings


def tensor_network_ag_compress_projector(
    tn: TensorNetwork,
    max_bond=None,
    cutoff=1e-10,
    site_tags=None,
    canonize=True,
    canonize_opts=None,
    lazy=False,
    optimize="auto-hq",
    equalize_norms=False,
    inplace=False,
    **compress_opts,
):
    """Compress an arbtrary geometry tensor network, with potentially multiple
    tensors per site, using locally computed projectors.

    Very loosely, this is like a generalization HOTRG.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compress. Every tensor should have exactly one of
        the site tags. Each site can have multiple tensors and output indices.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        A dynamic threshold for discarding singular values when compressing.
    site_tags : sequence of str, optional
        The tags to use to group the tensors from ``tn``. If not
        given, uses ``tn.site_tags``. The tensor network built will have one
        tensor per site.
    canonize : bool, optional
        Whether to pseudo canonicalize the initial tensor network.
    canonize_opts
        Supplied to :meth:`~quimb.tensor.tensor_core.TensorNetwork.gauge_all`.
    lazy : bool, optional
        Whether to leave the computed projectors uncontracted, default: False.
    optimize : str, optional
        The contraction path optimizer to use.
    equalize_norms : bool or float, optional
        Whether to equalize the norms of the tensors after compression. If an
        explicit value is give, then the norms will be set to that value, and
        the overall scaling factor will be accumulated into `.exponent`.
    inplace : bool, optional
        Whether to perform the compression inplace.
    compress_opts
        Supplied to :func:`~quimb.tensor.tensor_split`.

    Returns
    -------
    TensorNetwork
    """
    tn = tn if inplace else tn.copy()

    if site_tags is None:
        site_tags = tn.site_tags

    edges, _ = create_lazy_edge_map(tn, site_tags)

    if canonize:
        # optionally precondition the uncontracted network
        canonize_opts = ensure_dict(canonize_opts)
        gauges = canonize_opts.pop("gauges", {})

        if canonize == "layered":
            # get all tids for a single site
            tid0s = tn._get_tids_from_tags(site_tags[0])
            # get a tree span out of this region, and then group into layers
            # depending on which initial region tid they are connected to
            groups = [{tid} for tid in tid0s]
            for tida, tidb, _ in tn.get_tree_span(tid0s, inwards=False):
                next(g for g in groups if tidb in g).add(tida)

            # select each layer and gauge it separately
            tns = [tn._select_tids(group) for group in groups]
            for stn in tns:
                stn.gauge_all_simple_(gauges=gauges, **canonize_opts)

        else:
            # # global gauge
            tn.gauge_all_simple_(gauges=gauges, **canonize_opts)

        tn_calc = tn.copy()
        # have to insert gauges back into target before we insert projectors
        tn.gauge_simple_insert(gauges)
    else:
        tn_calc = tn.copy()
        gauges = None

    # then compute projectors using local information

    for taga, tagb in edges:
        #      │         │
        #    ──O─┐ chi ┌─O──
        #      │ └─▷═◁─┘ │
        #      │ ┌┘   └┐ │
        #    ──O─┘     └─O──
        #    taga       tagb
        tn_calc.insert_compressor_between_regions_(
            [taga],
            [tagb],
            max_bond=max_bond,
            cutoff=cutoff,
            insert_into=tn,
            new_ltags=[taga],
            new_rtags=[tagb],
            gauges=gauges,
            optimize=optimize,
            **compress_opts,
        )

    if not lazy:
        # then contract each site with all surrounding projectors
        for st in site_tags:
            tn.contract_(st, optimize=optimize)

    # XXX: do better than simply waiting til the end to equalize norms
    if equalize_norms is True:
        tn.equalize_norms_()
    elif equalize_norms:
        tn.equalize_norms_(value=equalize_norms)

    return tn


def tensor_network_ag_compress_local_early(
    tn: TensorNetwork,
    max_bond=None,
    cutoff=1e-10,
    site_tags=None,
    canonize=True,
    tree_gauge_distance=None,
    canonize_distance=None,
    canonize_after_distance=None,
    mode="auto",
    optimize="auto-hq",
    equalize_norms=False,
    inplace=False,
    **compress_opts,
):
    """Compress an arbtrary geometry tensor network, with potentially multiple
    tensors per site, using explicit contraction followed by immediate
    ('early') compression. In other words, contractions are interleaved with
    compressions.

    Very loosely, this is like a generalization of the 'zip-up' algorithm in
    1D, but for arbitrary geometry.


    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compress. Every tensor should have exactly one of
        the site tags. Each site can have multiple tensors and output indices.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        A dynamic threshold for discarding singular values when compressing.
    site_tags : sequence of str, optional
        The tags to use to group the tensors from ``tn``. If not
        given, uses ``tn.site_tags``. The tensor network built will have one
        tensor per site.
    canonize : bool, optional
        Whether to locally gauge before each compression, defaults to True.
    tree_gauge_distance : int, optional
        The distance to locally gauge to before each compression. Defaults to
        3.
    canonize_distance : int, optional
        The distance to canonize to before each compression, by default this
        is set by ``tree_gauge_distance``.
    canonize_after_distance : int, optional
        The distance to canonize to after each compression, by default this
        is set by ``tree_gauge_distance``, depending on ``mode``.
    mode : {'auto', 'basic', 'virtual-tree', ...}, optional
        The mode to use for the local gauging. If 'auto' will default to
        virtual tree gauging, or basic if `tree_gauge_distance` is 0.
    optimize : str, optional
        The contraction path optimizer to use.
    equalize_norms : bool or float, optional
        Whether to equalize the norms of the tensors after compression. If an
        explicit value is give, then the norms will be set to that value, and
        the overall scaling factor will be accumulated into `.exponent`.
    inplace : bool, optional
        Whether to perform the compression inplace.
    compress_opts
        Supplied to
        :meth:`~quimb.tensor.tensor_core.TensorNetwork.compress_between`.

    Returns
    -------
    TensorNetwork
    """
    tnc = tn if inplace else tn.copy()

    if site_tags is None:
        site_tags = tnc.site_tags

    _, neighbors = create_lazy_edge_map(tnc, site_tags)

    canonize_distance, canonize_after_distance, mode = (
        choose_local_compress_gauge_settings(
            canonize,
            tree_gauge_distance,
            canonize_distance,
            canonize_after_distance,
            mode,
        )
    )

    st0 = next(iter(site_tags))
    seen = {st0}
    queue = [st0]

    while queue:
        # process sites in a breadth-first manner
        taga = queue.pop(0)

        for tagb in neighbors[taga]:
            if tagb not in seen:
                queue.append(tagb)
                seen.add(tagb)

        # contract this site
        tnc.contract_(taga, optimize=optimize)

        # then immediately compress around it
        (tida,) = tnc._get_tids_from_tags(taga)
        for tidb in tnc._get_neighbor_tids(tida):
            tnc._compress_between_tids(
                tida,
                tidb,
                max_bond=max_bond,
                cutoff=cutoff,
                canonize_distance=canonize_distance,
                canonize_after_distance=canonize_after_distance,
                mode=mode,
                **compress_opts,
            )

    if equalize_norms is True:
        tnc.equalize_norms_()
    elif equalize_norms:
        tnc.equalize_norms_(value=equalize_norms)

    return tnc


def tensor_network_ag_compress_local_late(
    tn: TensorNetwork,
    max_bond=None,
    cutoff=1e-10,
    site_tags=None,
    canonize=True,
    tree_gauge_distance=None,
    canonize_distance=None,
    canonize_after_distance=None,
    mode="auto",
    optimize="auto-hq",
    equalize_norms=False,
    inplace=False,
    **compress_opts,
):
    """Compress an arbtrary geometry tensor network, with potentially multiple
    tensors per site, by explicitly contracting all sites first and then
    ('late') locally compressing. In other words, all contractions happen, then
    all compressions happen.

    Very loosely, this is like a generalization of the 'direct' algorithm in
    1D, but for arbitrary geometry.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compress. Every tensor should have exactly one of
        the site tags. Each site can have multiple tensors and output indices.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        A dynamic threshold for discarding singular values when compressing.
    site_tags : sequence of str, optional
        The tags to use to group the tensors from ``tn``. If not
        given, uses ``tn.site_tags``. The tensor network built will have one
        tensor per site.
    canonize : bool, optional
        Whether to locally gauge before each compression, defaults to True.
    tree_gauge_distance : int, optional
        The distance to locally gauge to before each compression. Defaults to
        3.
    canonize_distance : int, optional
        The distance to canonize to before each compression, by default this
        is set by ``tree_gauge_distance``.
    canonize_after_distance : int, optional
        The distance to canonize to after each compression, by default this
        is set by ``tree_gauge_distance``, depending on ``mode``.
    mode : {'auto', 'basic', 'virtual-tree', ...}, optional
        The mode to use for the local gauging. If 'auto' will default to
        virtual tree gauging, or basic if `tree_gauge_distance` is 0.
    optimize : str, optional
        The contraction path optimizer to use.
    equalize_norms : bool or float, optional
        Whether to equalize the norms of the tensors after compression. If an
        explicit value is give, then the norms will be set to that value, and
        the overall scaling factor will be accumulated into `.exponent`.
    inplace : bool, optional
        Whether to perform the compression inplace.
    compress_opts
        Supplied to
        :meth:`~quimb.tensor.tensor_core.TensorNetwork.compress_between`.

    Returns
    -------
    TensorNetwork
    """
    tnc = tn if inplace else tn.copy()

    if site_tags is None:
        site_tags = tnc.site_tags

    for st in site_tags:
        tnc.contract_(st, optimize=optimize)

    tnc.compress_all_(
        max_bond=max_bond,
        cutoff=cutoff,
        canonize=canonize,
        tree_gauge_distance=tree_gauge_distance,
        canonize_distance=canonize_distance,
        canonize_after_distance=canonize_after_distance,
        mode=mode,
        **compress_opts,
    )

    if equalize_norms is True:
        tnc.equalize_norms_()
    elif equalize_norms:
        tnc.equalize_norms_(value=equalize_norms)

    return tnc


def tensor_network_ag_compress_superorthogonal(
    tn: TensorNetwork,
    max_bond=None,
    cutoff=1e-10,
    site_tags=None,
    canonize=True,
    optimize="auto-hq",
    equalize_norms=False,
    inplace=False,
    **compress_opts,
):
    """Compress an arbtrary geometry tensor network, with potentially multiple
    tensors per site, using the 'superorthogonal' / 'Vidal' / quasi-canonical
    / 'simple update' gauge for compression. This is the same gauge as used in
    L2BP, but the intermediate tensor network is explicitly constructed.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compress. Every tensor should have exactly one of
        the site tags. Each site can have multiple tensors and output indices.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        A dynamic threshold for discarding singular values when compressing.
    site_tags : sequence of str, optional
        The tags to use to group the tensors from ``tn``. If not
        given, uses ``tn.site_tags``. The tensor network built will have one
        tensor per site.
    canonize : bool, optional
        Whether to locally gauge before each compression, defaults to True.
    optimize : str, optional
        The contraction path optimizer to use.
    equalize_norms : bool or float, optional
        Whether to equalize the norms of the tensors after compression. If an
        explicit value is give, then the norms will be set to that value, and
        the overall scaling factor will be accumulated into `.exponent`.
    inplace : bool, optional
        Whether to perform the compression inplace.
    compress_opts
        Supplied to
        :meth:`~quimb.tensor.tensor_core.TensorNetwork.compress_all_simple`.

    Returns
    -------
    TensorNetwork
    """
    tnc = tn if inplace else tn.copy()

    if site_tags is None:
        site_tags = tnc.site_tags

    for st in site_tags:
        tnc.contract_(st, optimize=optimize)

    tnc.fuse_multibonds_()

    if not canonize:
        # turn off gauging effect
        compress_opts.setdefault("max_iterations", 1)
        compress_opts.setdefault("tol", 0.0)
    else:
        compress_opts.setdefault("max_iterations", 1000)
        compress_opts.setdefault("tol", 5e-6)

    tnc.compress_all_simple_(
        max_bond=max_bond,
        cutoff=cutoff,
        **compress_opts,
    )

    if equalize_norms is True:
        tnc.equalize_norms_()
    elif equalize_norms:
        tnc.equalize_norms_(value=equalize_norms)

    return tnc


def tensor_network_ag_compress_l2bp(
    tn: TensorNetwork,
    max_bond=None,
    cutoff=1e-10,
    site_tags=None,
    canonize=True,
    damping=0.0,
    local_convergence=True,
    update="sequential",
    optimize="auto-hq",
    equalize_norms=False,
    inplace=False,
    **compress_opts,
):
    """Compress an arbitrary geometry tensor network, with potentially multiple
    tensors per site, using lazy 2-norm belief propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compress. Every tensor should have exactly one of
        the site tags. Each site can have multiple tensors and output indices.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        A dynamic threshold for discarding singular values when compressing.
    site_tags : sequence of str, optional
        The tags to use to group the tensors from ``tn``. If not
        given, uses ``tn.site_tags``. The tensor network built will have one
        tensor per site.
    canonize : bool, optional
        Whether to locally gauge before each compression, defaults to True.
    damping : float, optional
        How much to dampen message updates, to help convergence, defaults to 0.
    local_convergence : bool, optional
        Whether to use local convergence criteria, defaults to True.
    update : {'parallel', 'sequential'}, optional
        Whether to update all messages in parallel or sequentially, defaults to
        'parallel'.
    optimize : str, optional
        The contraction path optimizer to use.
    equalize_norms : bool or float, optional
        Whether to equalize the norms of the tensors after compression. If an
        explicit value is give, then the norms will be set to that value, and
        the overall scaling factor will be accumulated into `.exponent`.
    inplace : bool, optional
        Whether to perform the compression inplace.
    compress_opts
        Supplied to
        :func:`~quimb.tensor.belief_propagation.l2bp.compress_l2bp`.

    Returns
    -------
    TensorNetwork
    """
    from quimb.tensor.belief_propagation.l2bp import compress_l2bp

    if not canonize:
        compress_opts.setdefault("max_iterations", 1)

    tnc = compress_l2bp(
        tn,
        max_bond=max_bond,
        cutoff=cutoff,
        site_tags=site_tags,
        damping=damping,
        local_convergence=local_convergence,
        update=update,
        optimize=optimize,
        inplace=inplace,
        **compress_opts,
    )

    if equalize_norms is True:
        tnc.equalize_norms_()
    elif equalize_norms:
        tnc.equalize_norms_(value=equalize_norms)

    return tnc


_TNAG_COMPRESS_METHODS = {
    "local-early": tensor_network_ag_compress_local_early,
    "local-late": tensor_network_ag_compress_local_late,
    "projector": tensor_network_ag_compress_projector,
    "su": tensor_network_ag_compress_superorthogonal,
    "superorthogonal": tensor_network_ag_compress_superorthogonal,
    "l2bp": tensor_network_ag_compress_l2bp,
}


def tensor_network_ag_compress(
    tn: TensorNetwork,
    max_bond,
    cutoff=1e-10,
    method="local-early",
    site_tags=None,
    canonize=True,
    optimize="auto-hq",
    equalize_norms=False,
    inplace=False,
    **kwargs,
):
    """Compress an arbitrary geometry tensor network, with potentially multiple
    tensors per site.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compress. Every tensor should have exactly one of
        the site tags. Each site can have multiple tensors and output indices.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        A dynamic threshold for discarding singular values when compressing.
    method : {'local-early', 'local-late', 'projector', 'superorthogonal', 'l2bp'}, optional
        The compression method to use:

        - 'local-early': explicitly contract each site and interleave with
          immediate compression, see
          :func:`~quimb.tensor.tensor_arbgeom_compress.tensor_network_ag_compress_local_early`.
        - 'local-late': explicitly contract all sites and then compress, see
          :func:`~quimb.tensor.tensor_arbgeom_compress.tensor_network_ag_compress_local_late`.
        - 'projector': use locally computed projectors, see
          :func:`~quimb.tensor.tensor_arbgeom_compress.tensor_network_ag_compress_projector`.
        - 'superorthogonal': use the 'superorthogonal' gauge, see
          :func:`~quimb.tensor.tensor_arbgeom_compress.tensor_network_ag_compress_superorthogonal`.
        - 'l2bp': use lazy 2-norm belief propagation, see
          :func:`~quimb.tensor.tensor_arbgeom_compress.tensor_network_ag_compress_l2bp`.

    site_tags : sequence of str, optional
        The tags to use to group the tensors from ``tn``. If not
        given, uses ``tn.site_tags``. The tensor network built will have one
        tensor per site.
    canonize : bool, optional
        Whether to perform canonicalization, pseudo or otherwise depending on
        the method, before compressing.
    optimize : str, optional
        The contraction path optimizer to use.
    equalize_norms : bool or float, optional
        Whether to equalize the norms of the tensors after compression. If an
        explicit value is give, then the norms will be set to that value, and
        the overall scaling factor will be accumulated into `.exponent`.
    inplace : bool, optional
        Whether to perform the compression inplace.
    kwargs
        Supplied to the chosen compression method.
    """
    return _TNAG_COMPRESS_METHODS[method](
        tn,
        max_bond=max_bond,
        cutoff=cutoff,
        site_tags=site_tags,
        canonize=canonize,
        optimize=optimize,
        equalize_norms=equalize_norms,
        inplace=inplace,
        **kwargs,
    )



import functools
import math
import operator

import autoray as ar

from quimb.tensor import TensorNetwork, array_contract, bonds
from quimb.utils import RollingDiffMean


def prod(xs):
    """Product of all elements in ``xs``."""
    return functools.reduce(operator.mul, xs)


class BeliefPropagationCommon:
    """Common interfaces for belief propagation algorithms.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to perform belief propagation on.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    contract_every : int, optional
        If not None, 'contract' (via BP) the tensor network every
        ``contract_every`` iterations. The resulting values are stored in
        ``zvals`` at corresponding points ``zval_its``.
    callback : callable, optional
        A function to call after every iteration, of signature
        ``callback(bp_instance)``.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.
    """

    def __init__(
        self,
        tn: TensorNetwork,
        *,
        damping=0.0,
        update="sequential",
        normalize=None,
        distance=None,
        contract_every=None,
        callback=None,
        inplace=False,
    ):
        self.tn = tn if inplace else tn.copy()
        self.backend = self.tn.backend
        self.dtype = self.tn.dtype
        self.sign = 1.0
        self.exponent = tn.exponent
        self.damping = damping
        self.update = update
        self.callback = callback

        if normalize is None:
            if "complex" in self.dtype:
                normalize = "L2phased"
            else:
                normalize = "L2"
        self.normalize = normalize

        if distance is None:
            if ("complex" in self.dtype) and (
                callable(normalize) or ("phased" not in normalize)
            ):
                distance = "L2phased"
            else:
                distance = "L2"
        self.distance = distance

        self.contract_every = contract_every
        self.n = 0
        self.converged = False
        self.mdiffs = []
        self.rdiffs = []
        self.zval_its = []
        self.zvals = []

    @property
    def damping(self):
        return self._damping

    @damping.setter
    def damping(self, damping):
        if callable(damping):
            self._damping_fn = self._damping = damping
        else:
            self._damping = damping

            if damping == 0.0:

                def _damping_fn(old, new):
                    return new

            else:

                def _damping_fn(old, new):
                    return damping * old + (1 - damping) * new

            self._damping_fn = _damping_fn

    @property
    def normalize(self):
        return self._normalize

    @normalize.setter
    def normalize(self, normalize):
        if callable(normalize):
            # custom normalization function
            _normalize_fn = normalize

        elif normalize == "L1":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")

            def _normalize_fn(x):
                return x / _sum(_abs(x))

        elif normalize == "L2":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")

            def _normalize_fn(x):
                return x / (_sum(_abs(x) ** 2) ** 0.5)

        elif normalize == "L2phased":
            _sum = ar.get_lib_fn(self.backend, "sum")
            _abs = ar.get_lib_fn(self.backend, "abs")

            def _normalize_fn(x):
                xnrm = float(_sum(_abs(x) ** 2)) ** 0.5
                sumx = complex(_sum(x))
                if sumx != 0.0:
                    if sumx.imag == 0.0:
                        sumx = sumx.real
                    sumx /= abs(sumx)
                    xnrm *= sumx
                return x / xnrm

        elif normalize == "Linf":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _max = ar.get_lib_fn(self.backend, "max")

            def _normalize_fn(x):
                return x / _max(_abs(x))

        else:
            raise ValueError(f"Unrecognized normalize={normalize}")

        self._normalize = normalize
        self._normalize_fn = _normalize_fn

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, distance):
        if callable(distance):
            _distance_fn = distance

        elif distance == "L1":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")

            def _distance_fn(x, y):
                return float(_sum(_abs(x - y)))

        elif distance == "L2":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")

            def _distance_fn(x, y):
                return float(_sum(_abs(x - y) ** 2) ** 0.5)

        elif distance == "L2phased":
            _conj = ar.get_lib_fn(self.backend, "conj")
            _sum = ar.get_lib_fn(self.backend, "sum")
            _abs = ar.get_lib_fn(self.backend, "abs")

            def _distance_fn(x, y):
                xnorm = _sum(_abs(x) ** 2) ** 0.5
                ynorm = _sum(_abs(y) ** 2) ** 0.5
                # cosine similarity with phase
                cs = _sum(_conj(x) * y)
                phase = cs / _abs(cs)
                xn = x / xnorm
                yn = y / (ynorm * phase)
                # L2 distance between normalized, phased vectors
                return float(_sum(_abs(xn - yn) ** 2) ** 0.5)

        elif distance == "Linf":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _max = ar.get_lib_fn(self.backend, "max")

            def _distance_fn(x, y):
                return float(_max(_abs(x - y)))

        elif distance == "cosine":
            # this is like L2phased, but with less precision
            _conj = ar.get_lib_fn(self.backend, "conj")
            _sum = ar.get_lib_fn(self.backend, "sum")
            _abs = ar.get_lib_fn(self.backend, "abs")

            def _distance_fn(x, y):
                xnorm = float(_sum(_abs(x) ** 2) ** 0.5)
                ynorm = float(_sum(_abs(y) ** 2) ** 0.5)
                # compute cosine similarity
                cs = float(_abs(_sum(_conj(x) * y)) / (xnorm * ynorm))
                # clip to avoid numerical issues
                cs = min(max(cs, -1.0), 1.0)
                return (2 - 2 * cs) ** 0.5

        else:
            raise ValueError(f"Unrecognized distance={distance}")

        self._distance = distance
        self._distance_fn = _distance_fn

    def _maybe_contract(self):
        should_contract = (
            (self.contract_every is not None)
            and (self.n % self.contract_every == 0)
            and ((not self.zval_its) or (self.zval_its[-1] != self.n))
        )
        if should_contract:
            self.zval_its.append(self.n)
            self.zvals.append(self.contract())

    def run(
        self,
        max_iterations=1000,
        diis=False,
        tol=5e-6,
        tol_abs=None,
        tol_rolling_diff=None,
        info=None,
        progbar=False,
    ):
        """
        Parameters
        ----------
        max_iterations : int, optional
            The maximum number of iterations to perform.
        diis : bool or dict, optional
            Whether to use direct inversion in the iterative subspace to
            help converge the messages by extrapolating to low error guesses.
            If a dict, should contain options for the DIIS algorithm. The
            relevant options are {`max_history`, `beta`, `rcond`}.
        tol : float, optional
            The convergence tolerance for messages.
        tol_abs : float, optional
            The absolute convergence tolerance for maximum message update
            distance, if not given then taken as ``tol``.
        tol_rolling_diff : float, optional
            The rolling mean convergence tolerance for maximum message update
            distance, if not given then taken as ``tol``. This is used to stop
            running when the messages are just bouncing around the same level,
            without any overall upward or downward trends, roughly speaking.
        info : dict, optional
            If supplied, the following information will be added to it:
            ``converged`` (bool), ``iterations`` (int), ``max_mdiff`` (float),
            ``rolling_abs_mean_diff`` (float).
        progbar : bool, optional
            Whether to show a progress bar.
        """
        if tol_abs is None:
            tol_abs = tol
        if tol_rolling_diff is None:
            tol_rolling_diff = tol

        if progbar:
            import tqdm

            pbar = tqdm.tqdm()
        else:
            pbar = None

        if diis:
            from .diis import DIIS

            if isinstance(diis, dict):
                self._diis = DIIS(**diis)
                diis = True
            else:
                self._diis = DIIS()
        else:
            self._diis = None

        it = 0
        rdm = RollingDiffMean()
        self.converged = False
        while not self.converged and it < max_iterations:
            self._maybe_contract()

            # perform a single iteration of BP
            # we supply tol here for use with local convergence
            result = self.iterate(tol=tol)

            if diis:
                # extrapolate new guess for messages
                self.messages = self._diis.update(self.messages)

            if isinstance(result, dict):
                max_mdiff = result.get("max_mdiff", float("inf"))
            else:
                max_mdiff = result
                result = dict()

            self.mdiffs.append(max_mdiff)

            if pbar is not None:
                msg = f"max|dM|={max_mdiff:.3g}"

                nconv = result.get("nconv", None)
                if nconv is not None:
                    ncheck = result.get("ncheck", None)
                    msg += f" nconv: {nconv}/{ncheck} "

                pbar.set_description(msg, refresh=False)
                pbar.update()

            # check covergence criteria
            self.converged |= max_mdiff < tol_abs
            if tol_rolling_diff > 0.0:
                # check rolling mean convergence
                rdm.update(max_mdiff)
                amd = rdm.absmeandiff()
                self.converged |= amd < tol_rolling_diff
                self.rdiffs.append(amd)

            it += 1
            self.n += 1

            if self.callback is not None:
                self.callback(self)

        self._maybe_contract()

        # finally:
        if pbar is not None:
            pbar.close()

        if tol != 0.0 and not self.converged:
            import warnings

            warnings.warn(
                f"Belief propagation did not converge after {max_iterations} "
                f"iterations, tol={tol:.2e}, max|dM|={max_mdiff:.2e}."
            )

        if info is not None:
            info["converged"] = self.converged
            info["iterations"] = it
            info["max_mdiff"] = max_mdiff
            info["rolling_abs_mean_diff"] = rdm.absmeandiff()

    def plot(self, zvals_yscale="asinh", **kwargs):
        from quimb import plot_multi_series_zoom

        data = {
            "zvals": {
                "x": self.zval_its,
                "y": self.zvals,
                "yscale": zvals_yscale,
            },
            "mdiffs": self.mdiffs,
            "rdiffs": self.rdiffs,
        }
        if getattr(self, "_diis", None) is not None:
            data["diis.lambdas"] = self._diis.lambdas

        kwargs.setdefault("yscale", "log")
        return plot_multi_series_zoom(data, **kwargs)

    @property
    def mdiff(self):
        try:
            return self.mdiffs[-1]
        except IndexError:
            return float("nan")

    def iterate(self, tol=1e-6):
        """Perform a single iteration of belief propagation. Subclasses should
        implement this method, returning either `max_mdiff` or a dictionary
        containing `max_mdiff` and any other relevant information:

            {
                "nconv": nconv,
                "ncheck": ncheck,
                "max_mdiff": max_mdiff,
            }

        """
        raise NotImplementedError

    def contract(
        self,
        strip_exponent=False,
        check_zero=True,
        **kwargs,
    ):
        """Contract the tensor network and return the resulting value."""
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}(n={self.n}, mdiff={self.mdiff:.3g})"


def initialize_hyper_messages(
    tn,
    fill_fn=None,
    smudge_factor=1e-12,
):
    """Initialize messages for belief propagation, this is equivalent to doing
    a single round of belief propagation with uniform messages.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to initialize messages for.
    fill_fn : callable, optional
        A function to fill the messages with, of signature ``fill_fn(shape)``.
    smudge_factor : float, optional
        A small number to add to the messages to avoid numerical issues.

    Returns
    -------
    messages : dict
        The initial messages. For every index and tensor id pair, there will
        be a message to and from with keys ``(ix, tid)`` and ``(tid, ix)``.
    """
    backend = ar.infer_backend(next(t.data for t in tn))
    _sum = ar.get_lib_fn(backend, "sum")

    messages = {}

    # compute first messages from tensors to indices
    for tid, t in tn.tensor_map.items():
        k_inputs = tuple(range(t.ndim))
        for i, ix in enumerate(t.inds):
            if fill_fn is None:
                # sum over all other indices to get initial message
                m = array_contract(
                    arrays=(t.data,),
                    inputs=(k_inputs,),
                    output=(i,),
                )
                # normalize and insert
                messages[tid, ix] = m / _sum(m)
            else:
                d = t.ind_size(ix)
                m = fill_fn((d,))
                messages[tid, ix] = m / _sum(m)

    # compute first messages from indices to tensors
    for ix, tids in tn.ind_map.items():
        ms = [messages[tid, ix] for tid in tids]
        mp = prod(ms)
        for mi, tid in zip(ms, tids):
            m = mp / (mi + smudge_factor)
            # normalize and insert
            messages[ix, tid] = m / _sum(m)

    return messages


def combine_local_contractions(
    values,
    backend=None,
    strip_exponent=False,
    check_zero=True,
    mantissa=None,
    exponent=None,
):
    """Combine a product of local contractions into a single value, avoiding
    overflow/underflow by accumulating the mantissa and exponent separately.

    Parameters
    ----------
    values : sequence of (scalar, int)
        The values to combine, each with a power to be raised to.
    backend : str, optional
        The backend to use. Infered from the first value if not given.
    strip_exponent : bool, optional
        Whether to return the mantissa and exponent separately.
    check_zero : bool, optional
        Whether to check for zero values and return zero early.
    mantissa : float, optional
        The initial mantissa to accumulate into.
    exponent : float, optional
        The initial exponent to accumulate into.

    Returns
    -------
    result : float or (float, float)
        The combined value, or the mantissa and exponent separately.
    """
    if mantissa is None:
        mantissa = 1.0
    if exponent is None:
        exponent = 0.0

    _abs = _log10 = None
    for x, power in values:
        if _abs is None:
            # lazily get functions
            if backend is None:
                backend = ar.infer_backend(x)
            _abs = ar.get_lib_fn(backend, "abs")
            _log10 = ar.get_lib_fn(backend, "log10")

        # factor into phase and magnitude
        x_mag = _abs(x)
        x_phase = x / x_mag

        if check_zero and (x_mag == 0.0):
            # checking explicitly avoids errors from taking log(0)
            if strip_exponent:
                return 0.0, 0.0
            else:
                return 0.0

        # accumulate the mantissa and exponent separately,
        # accounting for the local power / counting factor
        mantissa = mantissa * x_phase**power
        exponent = exponent + power * _log10(x_mag)

    if strip_exponent:
        return mantissa, exponent
    else:
        return mantissa * 10**exponent


def contract_hyper_messages(
    tn,
    messages,
    strip_exponent=False,
    check_zero=True,
    backend=None,
):
    """Estimate the contraction of ``tn`` given ``messages``, via the
    exponential of the Bethe free entropy.
    """
    zvals = []

    for tid, t in tn.tensor_map.items():
        if backend is None:
            backend = ar.infer_backend(t.data)

        arrays = [t.data]
        inputs = [range(t.ndim)]
        for i, ix in enumerate(t.inds):
            arrays.append(messages[ix, tid])
            inputs.append((i,))

            # local message overlap correction
            z = array_contract(
                (messages[tid, ix], messages[ix, tid]),
                inputs=((0,), (0,)),
                output=(),
            )
            zvals.append((z, -1))

        # local factor free entropy
        z = array_contract(arrays, inputs, output=())
        zvals.append((z, 1))

    for ix, tids in tn.ind_map.items():
        arrays = tuple(messages[tid, ix] for tid in tids)
        inputs = tuple((0,) for _ in tids)
        # local variable free entropy
        z = array_contract(arrays, inputs, output=())
        zvals.append((z, 1))

    return combine_local_contractions(
        zvals,
        backend=backend,
        strip_exponent=strip_exponent,
        check_zero=check_zero,
    )


def compute_index_marginal(tn, ind, messages):
    """Compute the marginal for a single index given ``messages``.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compute the marginal for.
    ind : int
        The index to compute the marginal for.
    messages : dict
        The messages to use, which should match ``tn``.

    Returns
    -------
    marginal : array_like
        The marginal probability distribution for the index ``ind``.
    """
    m = prod(messages[tid, ind] for tid in tn.ind_map[ind])
    return m / ar.do("sum", m)


def compute_tensor_marginal(tn, tid, messages):
    """Compute the marginal for the region surrounding a single tensor/factor
    given ``messages``.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compute the marginal for.
    tid : int
        The tensor id to compute the marginal for.
    messages : dict
        The messages to use, which should match ``tn``.

    Returns
    -------
    marginal : array_like
        The marginal probability distribution for the tensor/factor ``tid``.
    """
    t = tn.tensor_map[tid]

    output = tuple(range(t.ndim))
    inputs = [output]
    arrays = [t.data]

    for i, ix in enumerate(t.inds):
        mix = prod(
            messages[otid, ix] for otid in tn.ind_map[ix] if otid != tid
        )
        inputs.append((i,))
        arrays.append(mix)

    m = array_contract(
        arrays=arrays,
        inputs=inputs,
        output=output,
    )

    return m / ar.do("sum", m)


def compute_all_index_marginals_from_messages(tn, messages):
    """Compute all index marginals from belief propagation messages.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compute marginals for.
    messages : dict
        The belief propagation messages.

    Returns
    -------
    marginals : dict
        The marginals for each index.
    """
    return {ix: compute_index_marginal(tn, ix, messages) for ix in tn.ind_map}


def normalize_message_pair(mi, mj):
    """Normalize a pair of messages such that `<mi|mj> = 1` and
    `<mi|mi> = <mj|mj>` (but in general != 1).
    """
    nij = ar.do("abs", mi @ mj) ** 0.5
    nii = (mi @ mi) ** 0.25
    njj = (mj @ mj) ** 0.25
    return mi / (nij * nii / njj), mj / (nij * njj / nii)


def maybe_get_thread_pool(thread_pool):
    """Get a thread pool if requested."""
    if thread_pool is False:
        return None

    if thread_pool is True:
        import quimb as qu

        return qu.get_thread_pool()

    if isinstance(thread_pool, int):
        import quimb as qu

        return qu.get_thread_pool(thread_pool)

    return thread_pool


def create_lazy_community_edge_map(tn, site_tags=None, rank_simplify=True):
    """For lazy BP algorithms, create the data structures describing the
    effective graph of the lazily grouped 'sites' given by ``site_tags``.
    """
    if site_tags is None:
        site_tags = set(tn.site_tags)
    else:
        site_tags = set(site_tags)

    edges = {}
    neighbors = {}
    local_tns = {}
    touch_map = {}

    for ix in tn.ind_map:
        ts = tn._inds_get(ix)
        tags = {tag for t in ts for tag in t.tags if tag in site_tags}
        if len(tags) >= 2:
            i, j = tuple(sorted(tags))

            if (i, j) in edges:
                # already processed this edge
                continue

            # add to neighbor map
            neighbors.setdefault(i, []).append(j)
            neighbors.setdefault(j, []).append(i)

            # get local TNs and compute bonds between them,
            # rank simplify here also to prepare for contractions
            try:
                tn_i = local_tns[i]
            except KeyError:
                tn_i = local_tns[i] = tn.select(i, virtual=False)
                if rank_simplify:
                    tn_i.rank_simplify_()
            try:
                tn_j = local_tns[j]
            except KeyError:
                tn_j = local_tns[j] = tn.select(j, virtual=False)
                if rank_simplify:
                    tn_j.rank_simplify_()

            edges[i, j] = tuple(bonds(tn_i, tn_j))

    for i, j in edges:
        touch_map[(i, j)] = tuple((j, k) for k in neighbors[j] if k != i)
        touch_map[(j, i)] = tuple((i, k) for k in neighbors[i] if k != j)

    if len(local_tns) != len(site_tags):
        # handle potentially disconnected sites
        for i in sorted(site_tags):
            try:
                tn_i = local_tns[i] = tn.select(i, virtual=False)
                if rank_simplify:
                    tn_i.rank_simplify_()
            except KeyError:
                pass

    return edges, neighbors, local_tns, touch_map


def auto_add_indices(tn, regions):
    """Make sure all indices incident to any tensor in each region are
    included in the region.
    """
    new_regions = []
    for r in regions:
        new_r = set(r)
        tids = [x for x in new_r if isinstance(x, int)]
        for tid in tids:
            t = tn.tensor_map[tid]
            new_r.update(t.inds)
        new_regions.append(frozenset(new_r))
    return new_regions


def process_loop_series_expansion_weights(
    weights,
    mantissa=1.0,
    exponent=0.0,
    multi_excitation_correct=True,
    maxiter_correction=100,
    tol_correction=1e-14,
    strip_exponent=False,
    return_all=False,
):
    """Assuming a normalized BP fixed point, take a series of loop weights, and
    iteratively compute the free energy by requiring self-consistency with
    exponential suppression factors. See https://arxiv.org/abs/2409.03108.
    """
    # this is the single exictation approximation
    f_uncorrected = -sum(weights.values())

    if multi_excitation_correct:
        # iteratively compute a self consistent free energy
        fold = float("inf")
        f = f_uncorrected
        for _ in range(maxiter_correction):
            f = -sum(
                wl * math.exp(len(gloop) * f) for gloop, wl in weights.items()
            )
            if abs(f - fold) < tol_correction:
                break
            fold = f
    else:
        f = f_uncorrected

    if return_all:
        return {gloop: math.exp(len(gloop) * f) for gloop in weights}

    mantissa = mantissa * (1 - f)

    if strip_exponent:
        return mantissa, exponent

    return mantissa * 10**exponent



"""This script illustrates how quimb's various MPI modes work.

It can be run as:

python ex_mpi_modes_expm_evo.py
quimb-mpi-python ex_mpi_modes_expm_evo.py
quimb-mpi-python --syncro ex_mpi_modes_expm_evo.py

And will display slightly different output for each that explains the
three modes.
"""

import quimb as qu
from mpi4py import MPI

# Get some MPI information
comm = MPI.COMM_WORLD
rank, size = comm.Get_rank(), comm.Get_size()
print(f"I am worker {rank} of total {size} running main script...")


# setup a verbose version of the ham_heis constructor, and make it Lazy
n = 18
shape = (2**n, 2**n)

# this makes the function print some information when called
#     - in order to be pickled is has to be located in the main package
ham_heis_verbose = qu.utils.Verbosify(
    qu.ham_heis, highlight="ownership", mpi=True
)

H = qu.Lazy(ham_heis_verbose, n=n, sparse=True, shape=shape)

# random initial state
#     - must make sure all processes have the same seed to be pure
psi0 = qu.rand_ket(2**n, seed=42)

# evolve the system, processes split 'hard' work (slepc computations)
#     - should see each worker gets given a different ownership rows
#     - but all end up with the results.
evo = qu.Evolution(psi0, H, method="expm", expm_backend="slepc")
evo.update_to(5)


print(f"{rank}: I have final state norm {qu.expec(evo.pt, evo.pt)}")

# Now lets demonstrate using the MPI pool construct
pool = qu.get_mpi_pool()

dims = [2] * n
bsz = 5
logneg_subsys_verbose = qu.utils.Verbosify(
    qu.logneg_subsys, highlight="sysb", mpi=True
)

# each process only computes its own fraction of these
#     - should see each process calls logneg with different ``sysb``.
fs = [
    pool.submit(
        logneg_subsys_verbose,
        evo.pt,
        dims=dims,
        sysa=range(0, bsz),
        sysb=range(i, i + bsz),
    )
    for i in range(bsz, n - bsz)
]

# but then the results are comminucated to everyone
rs = [f.result() for f in fs]

print(f"{rank}: I have logneg results: {rs}")



"""Region graph functionality - for GBP and cluster expansions."""

import functools
import itertools


def cached_region_property(name):
    """Decorator for caching information about regions."""

    def wrapper(meth):
        @functools.wraps(meth)
        def getter(self, region):
            try:
                return self.info[region][name]
            except KeyError:
                region_info = self.info.setdefault(region, {})
                region_info[name] = value = meth(self, region)
                return value

        return getter

    return wrapper


class RegionGraph:
    """A graph of regions, where each region is a set of nodes. For generalized
    belief propagation or cluster expansion methods.

    Parameters
    ----------
    regions : Iterable[Sequence[Hashable]]
        Generating regions.
    autocomplete : bool, optional
        Whether to automatically add all intersecting sub-regions, to guarantee
        a complete region graph.
    autoprune : bool, optional
        Whether to automatically remove all regions with a count of zero.
    """

    def __init__(
        self,
        regions=(),
        autocomplete=True,
        autoprune=True,
    ):
        regions = tuple(map(frozenset, regions))
        self.base_nodes = frozenset.intersection(*regions)

        self.lookup = {}
        self.parents = {}
        self.children = {}
        self.info = {}

        for region in regions:
            self.add_region(region)
        if autocomplete:
            self.autocomplete()
        if autoprune:
            self.autoprune()

    def reset_info(self):
        """Remove all cached region properties."""
        self.info.clear()

    @property
    def regions(self):
        return tuple(self.children)

    def get_overlapping(self, region):
        """Get all regions that intersect with the given region."""
        region = frozenset(region)

        if self.base_nodes:
            # all regions are overlapping
            return {
                other_region
                for other_region in self.regions
                if other_region != region
            }
        else:
            return {
                other_region
                for node in region
                for other_region in self.lookup[node]
                if other_region != region
            }

    def add_region(self, region):
        """Add a new region and update parent-child relationships.

        Parameters
        ----------
        region : Sequence[Hashable]
            The new region to add.
        """
        region = frozenset(region)

        if region in self.parents:
            # already added
            return

        # populate data structures
        self.parents[region] = set()
        self.children[region] = set()
        for node in region:
            # collect regions that contain nodes for fast neighbor lookup
            self.lookup.setdefault(node, set()).add(region)

        # add parent-child relationships
        for other in self.get_overlapping(region):
            if region.issubset(other):
                self.parents[region].add(other)
                self.children[other].add(region)
            elif other.issubset(region):
                self.children[region].add(other)
                self.parents[other].add(region)

        # prune redundant parents and children
        children = sorted(self.children[region], key=len)
        for i, c in enumerate(children):
            if any(c.issubset(cc) for cc in children[i + 1 :]):
                # child is a subset of larger child -> remove
                self.children[region].remove(c)
                self.parents[c].remove(region)

        parents = sorted(self.parents[region], key=len, reverse=True)
        for i, p in enumerate(parents):
            if any(p.issuperset(pp) for pp in parents[i + 1 :]):
                # parent is a superset of smaller parent -> remove
                self.parents[region].remove(p)
                self.children[p].remove(region)

            for c in self.children[region]:
                if p.issuperset(c):
                    # parent is a superset of child -> ensure no link
                    self.parents[c].discard(p)
                    self.children[p].discard(c)

        self.reset_info()

    def remove_region(self, region):
        """Remove a region and update parent-child relationships."""
        # remove from lookup
        for node in region:
            self.lookup[node].remove(region)

        # remove from parents and children, joining those up
        parents = self.parents.pop(region)
        children = self.children.pop(region)
        for p in parents:
            self.children[p].remove(region)
            self.children[p].update(children)
        for c in children:
            self.parents[c].remove(region)
            self.parents[c].update(parents)

        self.reset_info()

    def autocomplete(self):
        """Add all missing intersecting sub-regions."""
        if self.base_nodes:
            # all regions are overlapping
            rs = self.regions
            for i in range(len(rs)):
                for j in range(i + 1, len(rs)):
                    self.add_region(rs[i] & rs[j])
        else:
            for r in self.regions:
                for other in self.get_overlapping(r):
                    self.add_region(r & other)

    def autoprune(self):
        """Remove all regions with a count of zero."""
        for r in self.regions:
            if self.get_count(r) == 0:
                self.remove_region(r)

    def autoextend(self, regions=None):
        """Extend this region graph upwards by adding in all pairwise unions of
        regions. If regions is specified, take this as one set of pairs.
        """
        if regions is None:
            regions = self.regions

        neighbors = {}
        for r in regions:
            for other in self.get_overlapping(r):
                neighbors.setdefault(r, []).append(other)

        for r, others in neighbors.items():
            for other in others:
                self.add_region(r | other)

    def get_parents(self, region):
        """Get all ancestors that contain the given region, but do not contain
        any other regions that themselves contain the given region.
        """
        return self.parents[region]

    def get_children(self, region):
        """Get all regions that are contained by the given region, but are not
        contained by any other descendents of the given region.
        """
        return self.children[region]

    @cached_region_property("ancestors")
    def get_ancestors(self, region):
        """Get all regions that contain the given region, not just direct
        parents.
        """
        seen = set()
        queue = [region]
        while queue:
            r = queue.pop()
            for rp in self.parents[r]:
                if rp not in seen:
                    seen.add(rp)
                    queue.append(rp)
        return seen

    @cached_region_property("descendents")
    def get_descendents(self, region):
        """Get all regions that are contained by the given region, not just
        direct children.
        """
        seen = set()
        queue = [region]
        while queue:
            r = queue.pop()
            for rc in self.children[r]:
                if rc not in seen:
                    seen.add(rc)
                    queue.append(rc)
        return seen

    @cached_region_property("coparent_pairs")
    def get_coparent_pairs(self, region):
        """Get all regions which are direct parents of any descendant of the
        given region, but not themselves descendants of the given region.
        """
        # start with direct parents
        coparent_pairs = [(p, region) for p in self.get_parents(region)]

        # get all descendents
        rds = self.get_descendents(region)

        # exclude the region and its descendents
        seen = {region, *rds}

        # for each descendant
        for rd in rds:
            # add only its parents...
            for rdp in self.get_parents(rd):
                # ... which are not themselves descendents
                if rdp not in seen:
                    coparent_pairs.append((rdp, rd))
                    seen.add(rdp)

        return coparent_pairs

    @cached_region_property("count")
    def get_count(self, region):
        """Get the count of the given region, i.e. the correct weighting to
        apply when summing over all regions to avoid overcounting.
        """
        return 1 - sum(self.get_count(a) for a in self.get_ancestors(region))

    def get_total_count(self):
        """Get the total count of all regions."""
        return sum(map(self.get_count, self.regions))

    @cached_region_property("level")
    def get_level(self, region):
        """Get the level of the given region, i.e. the distance to an ancestor
        with no parents.
        """
        if not self.parents[region]:
            return 0
        else:
            return min(self.get_level(p) for p in self.get_parents(region)) - 1

    @cached_region_property("message_parts")
    def get_message_parts(self, pair):
        """Get the three contribution groups for a GBP message from region
        `source` to region `target`. 1. The part of region `source` that is
        not part of target, i.e. the factors to include. 2. The messages that
        appear in the numerator of the update equation. 3. The messages that
        appear in the denominator of the update equation.

        Parameters
        ----------
        source : Region
            The source region, should be a parent of `target`.
        target : Region
            The target region, should be a child of `source`.

        Returns
        -------
        factors : Region
            The difference of `source` and `target`, which will include the
            factors to appear in the numerator of the update equation.
        pairs_mul : set[(Region, Region)]
            The messages that appear in the numerator of the update equation,
            after cancelling out those that appear in the denominator.
        pairs_div : set[(Region, Region)]
            The messages that appear in the denominator of the update equation,
            after cancelling out those that appear in the numerator.
        """
        source, target = pair
        factors = source - target

        # we want to cancel out messages that appear in both of:
        # the messages that go into the belief of region `source`
        source_pairs = set(self.get_coparent_pairs(source))
        # the messages that go into the belief of region `target`
        target_pairs = set(self.get_coparent_pairs(target))
        # the current message to be updated by defn appears directly in the
        # update numerator, but also target belief region, so can be cancelled
        target_pairs.remove((source, target))

        pairs_mul = source_pairs - target_pairs
        pairs_div = target_pairs - source_pairs

        return factors, pairs_mul, pairs_div

    def check(self):
        """Run some basic consistency checks on the region graph."""
        for r, rps in self.parents.items():
            for rp in rps:
                assert r.issubset(rp)
                assert r in self.get_children(rp)

        for r in self.regions:
            for rd in self.get_descendents(r):
                assert r.issuperset(rd)
                assert r in self.get_ancestors(rd)

            for ra in self.get_ancestors(r):
                assert r.issubset(ra)
                assert r in self.get_descendents(ra)

            rps = self.get_parents(r)
            for rpa, rpb in itertools.combinations(rps, 2):
                assert not rpa.issubset(rpb)
                assert not rpb.issubset(rpa)

            rcs = self.get_children(r)
            for rca, rcb in itertools.combinations(rcs, 2):
                assert not rca.issubset(rcb)
                assert not rcb.issubset(rca)

    def draw(self, pos=None, a=20, scale=1.0, radius=0.1, **drawing_opts):
        from quimb.schematic import Drawing, hash_to_color

        if pos is None:
            pos = {node: node for node in self.lookup}

        def get_draw_pos(coo):
            return tuple(scale * s for s in pos[coo])

        sizes = {len(r) for r in self.regions}
        levelmap = {s: i for i, s in enumerate(sorted(sizes))}
        centers = {}

        d = Drawing(a=a, **drawing_opts)
        for region in sorted(self.regions, key=len, reverse=True):
            # level = self.get_level(region)
            # level = len(region)
            level = levelmap[len(region)]

            coos = [(*get_draw_pos(coo), 2.0 * level) for coo in region]

            average_coo = tuple(map(sum, zip(*coos)))
            centers[region] = tuple(c / len(coos) for c in average_coo)

            d.patch_around(
                coos,
                radius=radius,
                # edgecolor=hash_to_color(str(region)),
                facecolor=hash_to_color(str(region)),
                alpha=1 / 3,
                linestyle="",
                linewidth=3,
            )

        for region in self.regions:
            for child in self.get_children(region):
                d.line(
                    centers[region],
                    centers[child],
                    linewidth=0.5,
                    linestyle="-",
                    color=(0.5, 0.5, 0.5),
                    alpha=0.5,
                    arrowhead={},
                )

        return d.fig, d.ax

    def __repr__(self):
        return (
            f"<RegionGraph(regions={len(self.regions)}, "
            f"total_count={self.get_total_count()})>"
        )


def trie_add(trie, region):
    """Add a region to a trie data structure. This will create a nested
    dictionary structure where each node in the region corresponds to a
    key in the dictionary. The final leaf node will have a special key
    "__leaf__" that contains the original region.
    """
    try:
        canonical_ordered = sorted(region)
    except TypeError:
        canonical_ordered = sorted(region, key=repr)

    t = trie
    for node in canonical_ordered:
        t = t.setdefault(node, {})
    t["__leaf__"] = region


def trie_gen_subsets(trie, region):
    """Given a trie of regions, lazily generate all strict subsets
    (descendants) of the given region.
    """
    regionset = set(region)
    queue = [trie]
    while queue:
        t = queue.pop()
        rsub = t.get("__leaf__", None)
        if rsub is not None and rsub != region:
            yield rsub
        queue.extend(t for node, t in t.items() if node in regionset)


def gen_region_counts(regions, autocomplete=True, autoprune=True):
    """Lazily generate all intersecting regions and their counts. This is
    faster that constructing the full RegionGraph, using a trie to get
    all subsets (descendants) of each region.

    Parameters
    ----------
    regions : Iterable[Sequence[Hashable]]
        Generating regions.
    autocomplete : bool, optional
        Whether to automatically add all intersecting sub-regions, to guarantee
        a complete region graph.
    autoprune : bool, optional
        Whether to automatically remove all regions with a count of zero.
        Default is True.

    Yields
    ------
    region : frozenset
        The region for which the count is being calculated.
    count : int
        The count of the region, which is the correct weighting to apply when
        summing over all regions to avoid overcounting.
    """
    # remove redundant nodes that appear in all regions
    # (we just re-add these when yielding)
    regions = tuple(map(frozenset, regions))

    if not regions:
        return

    base_region = frozenset.intersection(*regions)
    if base_region:
        regions = tuple(r - base_region for r in regions)

    counts = {}
    # for intersections
    lookup = {}
    # for subsets
    trie = {}

    for ri in regions:
        counts[ri] = 0
        trie_add(trie, ri)

        if autocomplete:
            # simultenously get previous overlapping regions ...
            rjs = set()
            for node in ri:
                others = lookup.setdefault(node, [])
                for rj in others:
                    rjs.add(rj)
                # ... as well as then inserting current region
                others.append(ri)

            # find intersections with all previous overlapping
            for rj in rjs:
                rij = ri & rj
                if rij not in counts:
                    counts[rij] = 0
                    trie_add(trie, rij)

    # by ordering by size, we ensure visiting all ancestors first
    for ri in sorted(counts, key=len, reverse=True):
        # finalize current level (1 - count of all ancestors)
        ci = 1 - counts[ri]
        if ci == 0 and autoprune:
            # redundant node, contributes no more
            continue

        yield ri | base_region, ci

        for rj in trie_gen_subsets(trie, ri):
            counts[rj] += ci



import pytest
import autoray as ar

import quimb as qu
import quimb.tensor as qtn


class TestTensorNetwork3D:
    def test_cyclic_basic(self):
        tn = qtn.TN3D_empty(Lx=3, Ly=4, Lz=5, D=2, cyclic=True)
        assert tn.is_cyclic_x()
        assert tn.is_cyclic_y()
        assert tn.is_cyclic_z()
        assert tn.num_indices == 3 * tn.nsites
        tn = qtn.TN3D_empty(Lx=3, Ly=4, Lz=5, D=2, cyclic=(False, False, True))
        assert not tn.is_cyclic_x()
        assert not tn.is_cyclic_y()
        assert tn.is_cyclic_z()
        assert tn.num_indices == 3 * tn.nsites - (tn.Lx * tn.Lz) - (
            tn.Ly * tn.Lz
        )
        tn = qtn.TN3D_empty(Lx=3, Ly=4, Lz=5, D=2, cyclic=(False, True, False))
        assert not tn.is_cyclic_x()
        assert tn.is_cyclic_y()
        assert not tn.is_cyclic_z()
        assert tn.num_indices == 3 * tn.nsites - (tn.Lx * tn.Ly) - (
            tn.Ly * tn.Lz
        )
        tn = qtn.TN3D_empty(Lx=3, Ly=4, Lz=5, D=2, cyclic=(True, False, False))
        assert tn.is_cyclic_x()
        assert not tn.is_cyclic_y()
        assert not tn.is_cyclic_z()
        assert tn.num_indices == 3 * tn.nsites - (tn.Lx * tn.Ly) - (
            tn.Lx * tn.Lz
        )


class Test3DManualContract:
    @pytest.mark.parametrize("canonize", [False, True])
    def test_contract_boundary_ising_model(self, canonize):
        L = 5
        beta = 0.3
        fex = -2.7654417752878
        tn = qtn.TN3D_classical_ising_partition_function(L, L, L, beta=beta)
        Z = tn.contract_boundary(max_bond=8, canonize=canonize)
        f = -qu.log(Z) / (L**3 * beta)
        assert f == pytest.approx(fex, rel=1e-3)

    @pytest.mark.parametrize("dims", [(10, 4, 3), (4, 3, 10), (3, 10, 4)])
    def test_contract_boundary_stopping_criterion(self, dims):
        tn = qtn.TN3D_from_fill_fn(
            lambda shape: ar.lazy.Variable(shape=shape, backend="numpy"),
            *dims,
            D=2,
        )
        tn.contract_boundary_(
            4, cutoff=0.0, final_contract=False, progbar=True
        )
        assert tn.max_bond() == 4
        assert 32 <= tn.num_tensors <= 40

    @pytest.mark.parametrize("lazy", [False, True])
    def test_coarse_grain_basics(self, lazy):
        tn = qtn.TN3D_from_fill_fn(
            lambda shape: ar.lazy.Variable(shape, backend="numpy"),
            Lx=6,
            Ly=7,
            Lz=8,
            D=2,
        )
        tncg = tn.coarse_grain_hotrg("x", max_bond=3, cutoff=0.0, lazy=lazy)
        assert (tncg.Lx, tncg.Ly, tncg.Lz) == (3, 7, 8)
        assert not tncg.outer_inds()
        assert tncg.max_bond() == 3
        assert "I4,0,0" not in tncg.tag_map
        assert "X3" not in tncg.tag_map

        tncg = tn.coarse_grain_hotrg("y", max_bond=3, cutoff=0.0, lazy=lazy)
        assert (tncg.Lx, tncg.Ly, tncg.Lz) == (6, 4, 8)
        assert not tncg.outer_inds()
        assert tncg.max_bond() == 3
        assert "I0,5,0" not in tncg.tag_map
        assert "Y4" not in tncg.tag_map

        tncg = tn.coarse_grain_hotrg("z", max_bond=3, cutoff=0.0, lazy=lazy)
        assert (tncg.Lx, tncg.Ly, tncg.Lz) == (6, 7, 4)
        assert "I0,0,5" not in tncg.tag_map
        assert "Z4" not in tncg.tag_map

    def test_contract_hotrg_ising_model(self):
        L = 5
        beta = 0.3
        fex = -2.7654417752878
        tn = qtn.TN3D_classical_ising_partition_function(L, L, L, beta=beta)
        tn.contract_hotrg_(max_bond=4, progbar=True, equalize_norms=1.0)
        Z = tn.item() * 10**tn.exponent
        f = -qu.log(Z) / (L**3 * beta)
        assert f == pytest.approx(fex, rel=1e-2)

    @pytest.mark.parametrize("cyclicx", [False, True])
    @pytest.mark.parametrize("cyclicy", [False, True])
    @pytest.mark.parametrize("cyclicz", [False, True])
    @pytest.mark.parametrize("mode", ["hotrg", "ctmrg"])
    def test_contract_cyclic(self, cyclicx, cyclicy, cyclicz, mode):
        Lx, Ly, Lz = 3, 4, 5
        chi = 3
        tn = qtn.TN3D_from_fill_fn(
            lambda shape: ar.lazy.Variable(shape=shape, backend="numpy"),
            Lx,
            Ly,
            Lz,
            D=2,
            cyclic=(cyclicx, cyclicy, cyclicz),
        )
        if mode == "hotrg":
            lZ = tn.contract_hotrg(max_bond=chi, cutoff=0.0)
        elif mode == "ctmrg":
            lZ = tn.contract_ctmrg(max_bond=chi, cutoff=0.0)

        if any((cyclicx, cyclicy, cyclicz)):
            assert lZ.history_max_size() < 2**16
        else:
            assert lZ.history_max_size() < 2**13



from quimb.tensor.tensor_arbgeom import *


def virtual_bond_operator(
    self,
    ind,
    operators=None,
    flatten=False,
    rehearse=False,
    **contract_opts,
):
    operators = {} if operators is None else dict(operators)

    ket = self.copy()
    tida, tidb = sorted(ket.ind_map[ind])

    # cut the bond open and create dual
    ixkl, ixkr, ixbl, ixbr = (rand_uuid() for _ in range(4))
    ket._cut_between_tids(tida, tidb, ixkl, ixkr)
    bra = ket.H.reindex_({ixkl: ixbl, ixkr: ixbr})

    # optionally apply list of operators to ket only
    for where, G in operators.items():
        ket.gate_(G, where)

    # form partial trace
    overlap = ket | bra
    if flatten:
        for site in overlap.gen_sites_present():
            overlap ^= site

    output_inds = (ixbl, ixbr, ixkl, ixkr)
    if rehearse == "tree":
        return overlap.contraction_tree(
            optimize=contract_opts.get("optimize", None),
            output_inds=output_inds,
        )

    # contract to matrix that maps cut bra bond to cut ket bond
    return overlap.contract_compressed(
        output_inds=output_inds,
        **contract_opts,
    ).to_dense([ixbl, ixbr], [ixkl, ixkr])


def virtual_site_operator(
    self,
    site,
    operators=None,
    flatten=False,
    max_bond=None,
    rehearse=False,
    **contract_opts,
):
    operators = {} if operators is None else dict(operators)

    ket = self.copy()
    (tid,) = ket._get_tids_from_tags(site)
    t = ket.pop_tensor(tid)

    # once the tensor is removed, the physical site is gone - re-add later
    phys_ind = self.site_ind(site)
    k_ix = tuple(ix for ix in t.inds if ix != phys_ind)
    reindex = {ix: rand_uuid() for ix in k_ix}
    b_ix = tuple(reindex.values())
    bra = ket.conj().reindex_(reindex)

    if phys_ind not in t.inds:
        dangling = None
    elif site in operators:
        # missing physical bond is the operator
        dangling = operators.pop(site)
    else:
        # missing physical bond is just a identity
        d = t.ind_size(phys_ind)
        dangling = do("eye", d, dtype=t.data.dtype, like=t.data)

    # apply other operators to ket
    for where, G in operators.items():
        t = ket[where]
        t.gate_(G, ket.site_ind(where))

    # form partial trace
    overlap = ket | bra
    if flatten:
        for site in overlap.gen_sites_present():
            overlap ^= site
        overlap.fuse_multibonds_()

    output_inds = (*b_ix, *k_ix)
    if rehearse == "tree":
        return overlap.contraction_tree(
            optimize=contract_opts.get("optimize", None),
            output_inds=output_inds,
        )

    X = overlap.contract_compressed(
        max_bond=max_bond,
        output_inds=output_inds,
        **contract_opts,
    ).to_dense(b_ix, k_ix)

    if dangling is not None:
        X = do("kron", X, dangling)

    return X


def product_expectation_exact(
    self,
    Gs,
    where,
    optimize="auto-hq",
    normalized=True,
    rehearse=False,
    **contract_opts,
):
    k = self.copy()
    b = k.conj()

    for site, G in zip(where, Gs):
        k.gate_(G, (site,), contract=True)

    tn = b & k

    if rehearse:
        return handle_rehearse(rehearse, tn, optimize, output_inds=())

    expec = tn.contract(
        optimize=optimize,
        **contract_opts,
    )

    if normalized:
        k = self.copy()
        tn = b & k
        nfact = tn.contract(
            optimize=optimize,
            **contract_opts,
        )
        expec = expec / nfact

    return expec


def product_expectation_cluster(
    self,
    Gs,
    where,
    normalized=True,
    max_distance=0,
    fillin=False,
    gauges=None,
    optimize="auto",
    rehearse=False,
    **contract_opts,
):
    # select a local neighborhood of tensors
    tids = self._get_tids_from_tags(tuple(map(self.site_tag, where)), "any")

    k = self._select_local_tids(
        tids,
        max_distance=max_distance,
        fillin=fillin,
        virtual=False,
    )

    if gauges is not None:
        # gauge the region with simple update style bond gauges
        k.gauge_simple_insert(gauges)

    return k.product_expectation_exact(
        Gs=Gs,
        where=where,
        optimize=optimize,
        normalized=normalized,
        rehearse=rehearse,
        **contract_opts,
    )



import pytest
import numpy as np
import quimb as qu


def test_sgn_convention():
    from quimb.tensor.decomp import sgn

    assert sgn(1) == 1
    assert sgn(2.0) == 1
    assert sgn(-1) == -1
    assert sgn(-2.0) == -1
    assert sgn(0) == 1
    assert sgn(0.0) == 1
    assert sgn(0.0 + 0.0j) == 1
    assert sgn(1.0 + 2.0j) != 1
    assert sgn(1.0 + 2.0j) != -1
    assert abs(sgn(1.0 + 2.0j)) == pytest.approx(1)


@pytest.mark.parametrize(
    "dtype", ["float64", "float32", "complex128", "complex64"]
)
def test_qr_stabilized_sign_bug(dtype):
    from quimb.tensor.decomp import qr_stabilized

    for _ in range(10):
        Q = qu.rand_uni(4, dtype=dtype)
        R = qu.rand_matrix(4, dtype=dtype)

        # make R strictly upper triangular
        ii, jj = np.indices(R.shape)
        R[ii >= jj] = 0.0

        X = Q @ R
        Q2, _, R2 = qr_stabilized(X)

        assert abs(np.linalg.norm((Q2 @ R2) - X)) < (
            1e-12 if dtype in ("float64", "complex128") else 1e-6
        )
